import functools
import logging
import math
import os
import signal
import sys
import threading
import time
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import contextmanager
from typing import Any, Generator, Iterable, Optional, Set, Type, cast, Set, List
from urllib.parse import urlparse

import pendulum
from tornado import web
from tornado.ioloop import IOLoop
from tornado.httpserver import HTTPServer

from prefect import config
from prefect.client import Client
from prefect.engine.state import Failed, Submitted
from prefect.run_configs import RunConfig, UniversalRun
from prefect.serialization.state import StateSchema
from prefect.serialization.run_config import RunConfigSchema
from prefect.utilities.context import context
from prefect.utilities.graphql import GraphQLResult, with_args

ascii_name = r"""
 ____            __           _        _                    _
|  _ \ _ __ ___ / _| ___  ___| |_     / \   __ _  ___ _ __ | |_
| |_) | '__/ _ \ |_ / _ \/ __| __|   / _ \ / _` |/ _ \ '_ \| __|
|  __/| | |  __/  _|  __/ (__| |_   / ___ \ (_| |  __/ | | | |_
|_|   |_|  \___|_|  \___|\___|\__| /_/   \_\__, |\___|_| |_|\__|
                                           |___/
"""
# Event to notify agent process to start looking for available flow runs.
AGENT_WAKE_EVENT = threading.Event()


@contextmanager
def exit_handler(agent: "Agent") -> Generator:
    exit_event = threading.Event()

    def _exit_handler(*args: Any, **kwargs: Any) -> None:
        agent.logger.info("Keyboard interrupt received! Shutting down...")
        exit_event.set()
        AGENT_WAKE_EVENT.set()

    original = signal.getsignal(signal.SIGINT)
    try:
        signal.signal(signal.SIGINT, _exit_handler)
        yield exit_event
    except SystemExit:
        pass
    finally:
        signal.signal(signal.SIGINT, original)


class HealthHandler(web.RequestHandler):
    """Respond to /api/health"""

    def get(self) -> None:
        # Empty json blob, may add more info later
        self.write({})


class PokeHandler(web.RequestHandler):
    """Respond to /api/poke

    The handler is expected to be called by user to notify agent of available
    flow runs waiting for execution.
    """

    def get(self) -> None:
        # Wake up agent that might be waiting for interval loop to complete.
        AGENT_WAKE_EVENT.set()


class Agent:
    """
    Base class for Agents. Information on using the Prefect agents can be found at
    https://docs.prefect.io/orchestration/agents/overview.html

    This Agent class is a standard point for executing Flows through the Prefect API. It is meant to
    have subclasses which inherit functionality from this class. The only piece that the
    subclasses should implement is the `deploy_flows` function, which specifies how to run a
    Flow on the given platform. It is built in this way to keep Prefect API logic standard
    but allows for platform specific customizability.

    Args:
        - agent_config_id (str, optional): An optional agent configuration ID that can be used to set
            configuration based on an agent from a backend API. If set, all configuration values will be
            pulled from backend agent configuration. If not set, any manual kwargs will be used.
        - name (str, optional): An optional name to give this agent. Can also be set through
            the environment variable `PREFECT__CLOUD__AGENT__NAME`. Defaults to "agent"
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - env_vars (dict, optional): a dictionary of environment variables and values that will
            be set on each flow run that this agent submits for execution
        - max_polls (int, optional): maximum number of times the agent will poll the Prefect API
            for flow runs; defaults to infinite
        - agent_address (str, optional): Address to serve internal api at. Currently this is
            just health checks for use by an orchestration layer. Leave blank for no api server
            (default).
        - no_cloud_logs (bool, optional): Disable logging to a Prefect backend for this agent
            and all deployed flow runs
    """

    # Loop intervals for query sleep backoff
    _loop_intervals = {
        0: 0.25,
        1: 0.5,
        2: 1.0,
        3: 2.0,
        4: 4.0,
        5: 8.0,
        6: 10.0,
    }

    def __init__(
        self,
        agent_config_id: str = None,
        name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        max_polls: int = None,
        agent_address: str = None,
        no_cloud_logs: bool = None,
    ) -> None:
        # Auth with an API key will be loaded from the config or disk by the Client
        self.client = Client(api_server=config.cloud.api)

        self.agent_config_id = agent_config_id
        self._agent_config: Optional[dict] = None

        self.name = name or config.cloud.agent.get("name", "agent")
        self.labels = labels or list(config.cloud.agent.get("labels", []))
        self.env_vars = env_vars or config.cloud.agent.get("env_vars", dict())
        self.max_polls = max_polls

        if no_cloud_logs is None:  # Load from config if unset
            self.log_to_cloud = config.cloud.send_flow_run_logs
        else:
            self.log_to_cloud = not no_cloud_logs

        self.heartbeat_period = 60  # exposed for testing
        self.agent_address = agent_address or config.cloud.agent.get("agent_address")

        # These track background task objects so we can tear them down on exit
        self._api_server: Optional[HTTPServer] = None
        self._api_server_loop: Optional[IOLoop] = None
        self._api_server_thread: Optional[threading.Thread] = None
        self._heartbeat_thread: Optional[threading.Thread] = None

        # Create the default logger
        self.logger = self._get_logger()

        # Store a set of flows that are being submitted to prevent duplicate submissions
        self.submitting_flow_runs: Set[str] = set()

        # Log configuration options
        self.logger.debug(f"Environment variables: {[*self.env_vars]}")
        self.logger.debug(f"Max polls: {self.max_polls}")
        self.logger.debug(f"Agent address: {self.agent_address}")
        self.logger.debug(f"Log to Cloud: {self.log_to_cloud}")
        self.logger.debug(f"Prefect backend: {config.backend}")

    @property
    def flow_run_api_key(self) -> Optional[str]:
        """
        Get the API key that the flow run should use to authenticate with Cloud.

        Currently just returns the key that the agent is using to query for flow runs
        """
        return self.client.api_key

    def start(self) -> None:
        """
        The main entrypoint to the agent process. Sets up the agent then continuously
        polls for work to submit.

        This is the only method that should need to be called externally.
        """

        try:
            self._setup_api_connection()

            # Call subclass hook
            self.on_startup()

            # Print some nice startup logs
            self._show_startup_display()

            # Start background tasks
            self._start_heartbeat_thread()
            if self.agent_address:
                self._start_agent_api_server()

            # Enter the main loop checking for new flows
            with exit_handler(self) as exit_event:
                self._enter_work_polling_loop(exit_event)

        finally:

            self.on_shutdown()
            self._stop_agent_api_server()
            self._stop_heartbeat_thread()

    # Subclass hooks -------------------------------------------------------------------
    # -- These are intended to be defined by specific agent types

    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Invoked when a flow should be deployed for execution by this agent

        Must be implemented by a child class.

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment

        Raises:
            - ValueError: if deployment attempted on unsupported Storage type
        """
        raise NotImplementedError()

    def heartbeat(self) -> None:
        """
        Invoked by the heartbeat thread on a loop.

        A hook for child classes to implement.
        """
        pass

    def on_startup(self) -> None:
        """
        Invoked when the agent is starting up after verifying the connection to the API
        but before background tasks are created and work begins

        A hook for child classes to optionally implement.
        """
        pass

    def on_shutdown(self) -> None:
        """
        Invoked when the event loop is exiting and the agent is shutting down.

        A hook for child classes to optionally implement.
        """
        pass

    # Main work loop -------------------------------------------------------------------
    # We've now entered implementation details of the agent base. Subclasses should
    # not need to worry about how this works.

    def _enter_work_polling_loop(self, exit_event: threading.Event) -> None:
        """
        This method will loop until `exit_event` is set or `max_polls` is reached. The
        primary call is to `_submit_deploy_flow_run_jobs` which handles querying for
        and submitting ready flow runs. If there were flow runs to process, the loop
        will repeat immediately to attempt to pull all ready work quickly. If not, the
        loop will backoff with an increasing sleep.
        """
        backoff_index = 0
        remaining_polls = math.inf if self.max_polls is None else self.max_polls

        # the max workers default has changed in 3.8. For stable results the
        # default 3.8 behavior is elected here.
        max_workers = min(32, (os.cpu_count() or 1) + 4)
        self.logger.debug(
            f"Running thread pool with {max_workers} workers to handle flow deployment"
        )

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            while not exit_event.is_set() and remaining_polls:
                # Reset the event in case it was set by poke handler.
                AGENT_WAKE_EVENT.clear()

                submitted_flow_runs = self._submit_deploy_flow_run_jobs(
                    executor=executor,
                )

                # Flow runs were found, so start the next iteration immediately
                if submitted_flow_runs:
                    backoff_index = 0
                # Otherwise, add to the index (unless we are at the max already)
                elif backoff_index < max(self._loop_intervals.keys()):
                    backoff_index += 1

                remaining_polls -= 1

                self.logger.debug(
                    "Sleeping flow run poller for {} seconds...".format(
                        self._loop_intervals[backoff_index]
                    )
                )

                # Wait for loop interval timeout or agent to be poked by
                # external process before querying for flow runs again.
                AGENT_WAKE_EVENT.wait(timeout=self._loop_intervals[backoff_index])

    def _submit_deploy_flow_run_jobs(self, executor: "ThreadPoolExecutor") -> list:
        """
        - Queries for ready flow runs
        - Submits calls of `_deploy_flow_run(flow_run)` to the executor
        - Tracks submitted runs to prevent double deployments

        Args:
            - executor (ThreadPoolExecutor): the interface to submit flow deployments in
                background threads

        Returns:
            - A list of submitted flow runs
        """
        # Query for ready runs -- allowing for intermittent failures by handling
        # exceptions
        try:
            flow_run_ids = self._get_ready_flow_runs()
        except Exception:
            self.logger.error("Failed to query for ready flow runs", exc_info=True)
            return []

        # Get the flow run metadata necessary for deployment in bulk; this also filters
        # flow runs that may have been submitted by another agent in the meantime
        try:
            flow_runs = self._get_flow_run_metadata(flow_run_ids)
        except Exception:
            self.logger.error("Failed to query for flow run metadata", exc_info=True)
            return []

        # Deploy each flow run in parallel with the executor
        for flow_run in flow_runs:
            self.logger.debug(f"Submitting flow run {flow_run.id} for deployment...")
            executor.submit(self._deploy_flow_run, flow_run).add_done_callback(
                functools.partial(
                    self._deploy_flow_run_completed_callback, flow_run_id=flow_run.id
                )
            )
            self.submitting_flow_runs.add(flow_run.id)

        return flow_runs

    def _deploy_flow_run(
        self,
        flow_run: "GraphQLResult",
    ) -> None:
        """
        Deploy a flow run and update Cloud with the resulting deployment info.
        If any errors occur when submitting the flow run, capture the error and log to
        Cloud.

        Args:
            - flow_run (GraphQLResult): The specific flow run to deploy
        """

        # Deploy flow run and mark failed if any deployment error
        try:
            # Wait for the flow run's start time. The agent pre-fetches runs that may
            # not need to start until up to 10 seconds later so we need to wait to
            # prevent the flow from starting early
            #
            # `state.start_time` is used instead of `flow_run.scheduled_start_time` for
            # execution; `scheduled_start_time` is only to record the originally scheduled
            # start time of the flow run
            #
            # There are two possible states the flow run could be in at this point
            # - Scheduled - in this case the flow run state will have a start time
            # - Running - in this case the flow run state will not have a start time so we default to now
            flow_run_state = StateSchema().load(flow_run.serialized_state)
            start_time = getattr(flow_run_state, "start_time", pendulum.now())
            delay_seconds = max(0, (start_time - pendulum.now()).total_seconds())
            if delay_seconds:
                self.logger.debug(
                    f"Waiting {delay_seconds}s to deploy flow run {flow_run.id} on "
                    "time..."
                )
                time.sleep(delay_seconds)

            self.logger.info(
                f"Deploying flow run {flow_run.id} to execution environment..."
            )

            self._mark_flow_as_submitted(flow_run)

            # Call the main deployment hook
            deployment_info = self.deploy_flow(flow_run)

            self.logger.info(f"Completed deployment of flow run {flow_run.id}")

            self._safe_write_run_log(
                flow_run,
                message="Submitted for execution: {}".format(deployment_info),
                level="INFO",
            )

        except Exception as exc:
            # On exception, we'll mark this flow as failed

            # if first failure was a state update error, we don't want to try another
            # state update
            if "State update failed" in str(exc):
                self.logger.debug("Updating Flow Run state failed: {}".format(str(exc)))
                return

            # This is to match existing past behavior, I cannot imagine we would reach
            # this point with a flow run that has no id
            if not getattr(flow_run, "id"):
                self.logger.error("Flow run is missing an id.", exc_info=True)
                return

            self.logger.error(
                f"Exception encountered while deploying flow run {flow_run.id}",
                exc_info=True,
            )

            self._safe_write_run_log(
                flow_run,
                message=str(exc),
                level="ERROR",
            )
            self._mark_flow_as_failed(flow_run=flow_run, message=str(exc))

            self.logger.error(f"Deployment of {flow_run.id} aborted!")

    def _deploy_flow_run_completed_callback(
        self, _: "Future", flow_run_id: str
    ) -> None:
        """
        Called on completion of `_deploy_flow_run` regardless of success.
        Clears the `flow_run_id` from the submitted flow runs cached. No matter what
        the result of the deployment job was, if the flow is in the ready queue from
        the backend API this agent should be open to trying to deploy it again.

        Args:
            - _ (Future): required for future callbacks but unused
            - flow_run_id (str): the id of the flow run that the future represents.
        """
        self.submitting_flow_runs.remove(flow_run_id)

    # Background jobs ------------------------------------------------------------------
    # - Agent API server
    # - Heartbeat thread

    def _start_agent_api_server(self) -> None:
        if not self.agent_address:
            raise ValueError("Cannot run agent API without setting `agent_address`")

        parsed = urlparse(self.agent_address)
        if not parsed.port:
            raise ValueError("Must specify port in agent address")
        port = cast(int, parsed.port)
        hostname = parsed.hostname or ""
        app = web.Application(
            [("/api/health", HealthHandler), ("/api/poke", PokeHandler)]
        )

        def run() -> None:
            # Ensure there's an active event loop in this thread
            import asyncio

            try:
                asyncio.get_running_loop()
            except RuntimeError:
                asyncio.set_event_loop(asyncio.new_event_loop())

            self.logger.debug(
                f"Agent API server listening on port {self.agent_address}"
            )
            self._api_server = app.listen(port, address=hostname)
            self._api_server_loop = IOLoop.current()
            self._api_server_loop.start()  # type: ignore

        self._api_server_thread = threading.Thread(
            name="api-server", target=run, daemon=True
        )
        self._api_server_thread.start()

    def _stop_agent_api_server(self) -> None:
        if self._api_server is not None:
            self.logger.debug("Stopping agent API server")
            self._api_server.stop()

        if self._api_server_loop is not None:
            self.logger.debug("Stopping agent API server loop")

            def stop_server() -> None:
                try:
                    loop = cast(IOLoop, self._api_server_loop)
                    loop.stop()
                except Exception:
                    pass

            self._api_server_loop.add_callback(stop_server)

        if self._api_server_thread is not None:
            self.logger.debug("Joining agent API threads")
            # Give the server a small period to shutdown nicely, otherwise it
            # will terminate on exit anyway since it's a daemon thread.
            self._api_server_thread.join(timeout=1)

    def _start_heartbeat_thread(self) -> None:
        """
        Run a thread to send heartbeats to the backend API, should be called at `start`
        """

        def run() -> None:
            while True:
                try:
                    self.logger.debug("Sending agent heartbeat...")
                    self.heartbeat()
                except Exception:
                    self.logger.error(
                        "Error in agent heartbeat, will try again in %.1f seconds...",
                        self.heartbeat_period,
                        exc_info=True,
                    )
                else:
                    self.logger.debug(
                        "Heartbeat succesful! Sleeping for %.1f seconds...",
                        self.heartbeat_period,
                    )
                time.sleep(self.heartbeat_period)

        self._heartbeat_thread = threading.Thread(
            name="heartbeat", target=run, daemon=True
        )
        self._heartbeat_thread.start()

    def _stop_heartbeat_thread(self) -> None:
        """
        Stop the heartbeat thread, should be called at `cleanup`
        """
        if self._heartbeat_thread is not None:
            self.logger.debug("Stopping heartbeat thread...")
            self._heartbeat_thread.join(timeout=1)

    # Backend API queries --------------------------------------------------------------

    def _get_ready_flow_runs(self, prefetch_seconds: int = 10) -> Set[str]:
        """
        Query the Prefect API for flow runs in the 'ready' queue. Results from here
        should be filtered by '_get_flow_run_metadata' to prevent duplicate submissions
        across agents.

        Args:
            - prefetch_seconds: The number of seconds in the future to fetch runs for.
                This reduces the chance of a flow starting late due to a polling loop
                backoff and should be set to the largest loop interval.
                `_deploy_flow_run` will wait until the true scheduled start time before
                deploying the flow.

        Returns:
            - set: A set of flow run ids that are ready
        """
        self.logger.debug("Querying for ready flow runs...")

        # keep a copy of what was submitted running before the query as the thread
        # pool might modify the set in the meantime
        currently_submitting_flow_runs = self.submitting_flow_runs.copy()

        # Get scheduled flow runs from queue
        mutation = {
            "mutation($input: get_runs_in_queue_input!)": {
                "get_runs_in_queue(input: $input)": {"flow_run_ids"}
            }
        }

        now = pendulum.now("UTC")
        result = self.client.graphql(
            mutation,
            variables={
                "input": {
                    "before": now.add(seconds=prefetch_seconds).isoformat(),
                    "labels": list(self.labels),
                    "tenant_id": self.client.tenant_id,
                }
            },
        )

        # we queried all of the available flow runs, however, some may have already been pulled
        # by this agent and are in the process of being submitted in the background. We do not
        # want to act on these "duplicate" flow runs until we've been assured that the background
        # thread has attempted to submit the work (successful or otherwise).
        flow_run_ids = set(result.data.get_runs_in_queue.flow_run_ids)  # type: ignore

        if flow_run_ids:
            msg = f"Found {len(flow_run_ids)} ready flow run(s): {flow_run_ids}"
        else:
            msg = "No ready flow runs found."

        already_submitting = flow_run_ids & currently_submitting_flow_runs
        target_flow_run_ids = flow_run_ids - already_submitting

        if already_submitting:
            msg += " ({} already being submitted: {})".format(
                len(already_submitting), list(already_submitting)
            )

        self.logger.debug(msg)
        return target_flow_run_ids

    def _get_flow_run_metadata(
        self,
        flow_run_ids: Iterable[str],
    ) -> List["GraphQLResult"]:
        """
        Get metadata about a collection of flow run ids that the agent is preparing
        to submit

        This function will filter the flow runs to a collection where:

        - The flow run is in a 'Scheduled' state. This prevents flow runs that have
          been submitted by another agent from being submitted again.

        - The flow run is in another state, but has task runs in a 'Running' state
          scheduled to start now. This is for retries in which the flow run is placed
          back into the ready queue but is not in a Scheduled state.

        Args:
            flow_run_ids: Flow run ids to query (order will not be respected)
            start_time: Only

        Returns:
           List: Metadata per flow run sorted by scheduled start time (ascending)
        """
        if not flow_run_ids:
            return []

        flow_run_ids = list(flow_run_ids)
        self.logger.debug(f"Retrieving metadata for {len(flow_run_ids)} flow run(s)...")

        # This buffer allows flow runs to retry immediately in their own deployment
        # without the agent creating a second deployment
        retry_start_time_buffer = pendulum.now("UTC").subtract(seconds=3).isoformat()

        where = {
            # Only get flow runs in the requested set
            "id": {"_in": flow_run_ids},
            # and filter by the additional criteria...
            "_or": [
                # This flow run has not been taken by another agent
                {"state": {"_eq": "Scheduled"}},
                # Or, this flow run has been set to retry and has not been immediately
                # retried in its own process
                {
                    "state": {"_eq": "Running"},
                    "task_runs": {
                        "state_start_time": {"_lte": retry_start_time_buffer}
                    },
                },
            ],
        }

        query = {
            "query": {
                with_args("flow_run", {"where": where}): {
                    "id": True,
                    "version": True,
                    "state": True,
                    "serialized_state": True,
                    "parameters": True,
                    "scheduled_start_time": True,
                    "run_config": True,
                    "name": True,
                    "flow": {
                        "id",
                        "name",
                        "environment",
                        "storage",
                        "version",
                        "core_version",
                    },
                    # Collect and return task run metadata as well so the state can be
                    # updated in `_mark_flow_as_submitted`
                    with_args(
                        "task_runs",
                        {
                            "where": {
                                "state_start_time": {"_lte": retry_start_time_buffer}
                            }
                        },
                    ): {"id", "version", "task_id", "serialized_state"},
                }
            }
        }
        result = self.client.graphql(query)
        return sorted(
            result.data.flow_run,
            key=lambda flow_run: flow_run.serialized_state.get(
                "start_time", pendulum.now("utc").isoformat()
            ),
        )

    def _mark_flow_as_submitted(self, flow_run: GraphQLResult) -> None:
        """
        After a flow run is grabbed this function sets the state to Submitted so it
        won't be picked up by any other processes

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object
        """
        # Set flow run state to `Submitted` if it is currently `Scheduled`
        if StateSchema().load(flow_run.serialized_state).is_scheduled():

            self.logger.debug(
                f"Updating flow run {flow_run.id} state from Scheduled -> Submitted..."
            )
            self.client.set_flow_run_state(
                flow_run_id=flow_run.id,
                version=flow_run.version,
                state=Submitted(
                    message="Submitted for execution",
                    state=StateSchema().load(flow_run.serialized_state),
                ),
            )

        # Set task run states to `Submitted` if they are currently `Scheduled`
        task_runs_updated = 0
        for task_run in flow_run.task_runs:
            if StateSchema().load(task_run.serialized_state).is_scheduled():
                task_runs_updated += 1
                self.client.set_task_run_state(
                    task_run_id=task_run.id,
                    version=task_run.version,
                    state=Submitted(
                        message="Submitted for execution.",
                        state=StateSchema().load(task_run.serialized_state),
                    ),
                )
        if task_runs_updated:
            self.logger.debug(
                f"Updated {task_runs_updated} task runs states for flow run "
                f"{flow_run.id} from  Scheduled -> Submitted"
            )

    def _mark_flow_as_failed(self, flow_run: GraphQLResult, message: str) -> None:
        """
        Mark a flow run as `Failed`

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object
            - exc (Exception): An exception that was raised to use as the `Failed`
                message
        """
        self.logger.error(
            f"Updating flow run {flow_run.id} state to Failed...",
        )
        self.client.set_flow_run_state(
            flow_run_id=flow_run.id,
            version=flow_run.version,
            state=Failed(message=message),
        )

    def _get_run_config(
        self, flow_run: GraphQLResult, run_config_cls: Type[RunConfig]
    ) -> RunConfig:
        """
        Get a run_config for the flow, if present. The returned run config is always of
        type `run_config_cls`

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object
            - run_config_cls (Callable): The expected run-config class

        Returns:
            - RunConfig: The flow run's run-config or an instance of `run_config_cls`
        """
        # If the flow is using a run_config, load it
        if getattr(flow_run, "run_config", None) is not None:
            run_config = RunConfigSchema().load(flow_run.run_config)
            if isinstance(run_config, UniversalRun):
                # Convert to agent-specific run-config
                return run_config_cls(env=run_config.env, labels=run_config.labels)
            elif not isinstance(run_config, run_config_cls):
                msg = (
                    "Flow run %s has a `run_config` of type `%s`, only `%s` is supported"
                    % (flow_run.id, type(run_config).__name__, run_config_cls.__name__)
                )
                self.logger.error(msg)
                raise TypeError(msg)
            return run_config

        # Otherwise, return the default run_config
        return run_config_cls()

    def _safe_write_run_log(
        self, flow_run: GraphQLResult, message: str, level: str
    ) -> None:
        """
        Write a log to the backend API for the given flow run. If the flow run object
        does not have an id, this is a no-op.

        Args:
            flow_run: The flow run object returned by a GQL query
            message: The log message
            level: The log level
        """
        if getattr(flow_run, "id", None):
            self.client.write_run_logs(
                [
                    dict(
                        flow_run_id=flow_run.id,
                        name=self.name,
                        message=message,
                        level=level,
                    )
                ]
            )

    # Backend API connection -----------------------------------------------------------

    def _register_agent(self) -> str:
        """
        Register this agent with a backend API and retrieve the ID

        Returns:
            - The agent ID as a string
        """

        config_id_blub = (
            f" with config {self.agent_config_id}" if self.agent_config_id else ""
        )
        self.logger.info(f"Registering agent{config_id_blub}...")

        agent_id = self.client.register_agent(
            agent_type=type(self).__name__,
            name=self.name,
            labels=self.labels,  # type: ignore
            agent_config_id=self.agent_config_id,
        )

        self.logger.info("Registration successful!")
        self.logger.debug(f"Assigned agent id: {agent_id}")

        if self.agent_config_id:
            self.agent_config = self._retrieve_agent_config()

        return agent_id

    def _retrieve_agent_config(self) -> dict:
        """
        Retrieve the configuration of an agent if an agent ID is provided

        Returns:
            - dict: a dictionary of agent configuration
        """
        if not self.agent_config_id:
            raise ValueError(
                "Cannot retrieve agent config without setting `agent_config_id`"
            )

        agent_config = self.client.get_agent_config(self.agent_config_id)
        self.logger.debug(f"Loaded agent config {self.agent_config_id}: {agent_config}")
        return agent_config

    def _setup_api_connection(self) -> None:
        """
        Sets up the agent's connection to Cloud

        - Gets an agent_id and attaches it to the headers
        - Runs a test query to check for a good setup

        Raises:
            RuntimeError: On failed test query
        """

        # Register agent with backend API
        self.client.attach_headers({"X-PREFECT-AGENT-ID": self._register_agent()})

        self.logger.debug(f"Sending test query to API at {config.cloud.api!r}...")

        try:
            self.client.graphql(query="query { hello }")
            self.logger.debug("Test query successful!")
        except Exception as exc:
            raise RuntimeError(
                f"Error while contacting API at {config.cloud.api}"
            ) from exc

    # Utilities ------------------------------------------------------------------------

    def _show_startup_display(self) -> None:
        print(ascii_name)
        self.logger.info(f"Starting {type(self).__name__} with labels {self.labels}")
        self.logger.info(
            "Agent documentation can be found at "
            "https://docs.prefect.io/orchestration/"
        )
        self.logger.info("Waiting for flow runs...")

    def _get_logger(self) -> logging.Logger:
        """
        Create an agent logger based on config options
        """

        logger = logging.getLogger(self.name)
        logger.setLevel(config.cloud.agent.get("level"))

        # Ensure it has a stream handler
        if not any(
            isinstance(handler, logging.StreamHandler) for handler in logger.handlers
        ):
            ch = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter(context.config.logging.format)
            formatter.converter = time.gmtime  # type: ignore
            ch.setFormatter(formatter)
            logger.addHandler(ch)

        return logger


if __name__ == "__main__":
    Agent().start()



import warnings
from typing import Any

from prefect.executors import LocalExecutor as _LocalExecutor


class LocalExecutor(_LocalExecutor):
    def __new__(cls, *args: Any, **kwargs: Any) -> "LocalExecutor":
        warnings.warn(
            "prefect.engine.executors.LocalExecutor has been moved to "
            "`prefect.executors.LocalExecutor`, please update your imports",
            stacklevel=2,
        )
        return super().__new__(cls)



from typing import Iterable

from prefect.run_configs.base import RunConfig


class DockerRun(RunConfig):
    """Configure a flow-run to run as a Docker container.

    Args:
        - image (str, optional): The image to use
        - env (dict, optional): Additional environment variables to set in the
            container
        - labels (Iterable[str], optional): an iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work
        - ports (Iterable[int], optional): an iterable of ports to pass to the
            Docker Agent which are used to expose container ports.
        - host_config (dict, optional): A dictionary or runtime arguments to pass to
            the Docker Agent. See link below for options.
            https://docker-py.readthedocs.io/en/stable/api.html#docker.api.container.ContainerApiMixin.create_host_config

    Examples:

    Use the defaults set on the agent:

    ```python
    flow.run_config = DockerRun()
    ```

    Set an environment variable in the flow run process:

    ```python
    flow.run_config = DockerRun(env={"SOME_VAR": "value"})
    ```
    """

    def __init__(
        self,
        *,
        image: str = None,
        env: dict = None,
        labels: Iterable[str] = None,
        ports: Iterable[int] = None,
        host_config: dict = None,
    ) -> None:
        super().__init__(env=env, labels=labels)
        self.image = image
        self.host_config = host_config
        self.ports = ports



import click

from prefect.client import Client
from prefect.exceptions import ClientError
from prefect.utilities.graphql import with_args


@click.group(hidden=True)
def create():
    """
    Create commands that refer to mutations of Prefect API metadata.

    \b
    Usage:
        $ prefect create [OBJECT]

    \b
    Arguments:
        project    Create projects

    \b
    Examples:
        $ prefect create project "Hello, World!"
        Hello, World! created

    \b
        $ prefect create project "Hello, World!" --description "My description"
        Hello, World! created
    """


@create.command(hidden=True)
@click.argument("name", required=True)
@click.option("--description", "-d", help="Project description to create", hidden=True)
@click.option(
    "--skip-if-exists",
    is_flag=True,
    help="Skip creation if project already exists",
    hidden=True,
)
def project(name, description, skip_if_exists):
    """
    Create projects with the Prefect API that organize flows. Does nothing if
    the project already exists.

    \b
    Arguments:
        name                TEXT    The name of a project to create     [required]

    \b
    Options:
        --description, -d   TEXT    A project description
        --skip-if-exists            Optionally skip creation call if project already exists

    """
    if skip_if_exists:
        result = Client().graphql(
            query={
                "query": {
                    with_args("project", {"where": {"name": {"_eq": name}}}): {
                        "id": True
                    }
                }
            }
        )
        if result.data.project:
            click.secho("{} already exists".format(name), fg="green")
            return

    try:
        Client().create_project(project_name=name, project_description=description)
    except ClientError as exc:
        click.echo(f"{type(exc).__name__}: {exc}")
        click.secho("Error creating project", fg="red")
        return

    click.secho("{} created".format(name), fg="green")



from packaging.version import parse

from prefect.utilities.graphql import GraphQLResult


def get_flow_image(flow_run: GraphQLResult, default: str = None) -> str:
    """
    Retrieve the image to use for this flow run deployment.

    Args:
        - flow_run (GraphQLResult): A GraphQLResult flow run object
        - default (str, optional): A default image to use. If not specified,
            The `prefecthq/prefect` image corresponding with the flow's prefect
            version will be used.

    Returns:
        - str: a full image name to use for this flow run

    Raises:
        - ValueError: if deployment attempted on unsupported Storage type and `image` not
            present in environment metadata
    """
    from prefect.storage import Docker
    from prefect.serialization.storage import StorageSchema
    from prefect.serialization.run_config import RunConfigSchema

    has_run_config = getattr(flow_run, "run_config", None) is not None

    storage = StorageSchema().load(flow_run.flow.storage)

    # Precedence:
    # - Image on docker storage
    # - Image on run_config
    # - Provided default
    # - `prefecthq/prefect` for flow's core version

    if isinstance(storage, Docker):
        return storage.name

    if has_run_config:
        run_config = RunConfigSchema().load(flow_run.run_config)
        if getattr(run_config, "image", None) is not None:
            return run_config.image

    if default is not None:
        return default

    # core_version should always be present, but just in case
    version = flow_run.flow.get("core_version") or "latest"
    cleaned_version = version.split("+")[0]

    return f"prefecthq/prefect:{cleaned_version}"


def get_flow_run_command(flow_run: GraphQLResult) -> str:
    """
    Determine the flow run command to use based on a flow's version. This is due to a command
    deprecation in `0.13.0` where `execute cloud-flow` was changes to `execute flow-run`

    Args:
        - flow_run (GraphQLResult): A GraphQLResult flow run object

    Returns:
        - str: a prefect CLI command to execute a flow run
    """
    core_version = getattr(flow_run.flow, "core_version", None) or "0.0.0"

    if parse(core_version) < parse("0.13.0"):
        return "prefect execute cloud-flow"

    return "prefect execute flow-run"



"""
Tasks for creating and inspecting Prefect flow runs


Example:
    ```python
    import prefect
    from prefect import task, Flow, Parameter
    from prefect.tasks.prefect.flow_run import (
        create_flow_run,
        get_task_run_result,
    )


    @task
    def create_some_data(length: int):
        return list(range(length))


    with Flow("child") as child_flow:
        data_size = Parameter("data_size", default=5)
        data = create_some_data(data_size)


    with Flow("parent") as parent_flow:
        child_run_id = create_flow_run(
            flow_name=child_flow.name, parameters=dict(data_size=10)
        )
        child_data = get_task_run_result(child_run_id, "create_some_data-1")

    if __name__ == "__main__":
        # Register the sub flow
        child_flow.register("default")

        # Run the parent flow locally
        flow_run = parent_flow.run()

        # Print the retrieved data
        print(flow_run.result[child_data].result)

    ```
"""

import datetime
import time
import pendulum
from typing import Any, Iterable, Optional, Union
from datetime import timedelta
from urllib.parse import urlparse

import prefect
from prefect import Client, Task, task
from prefect.backend.artifacts import create_link_artifact
from prefect.backend.flow_run import FlowRunView, FlowView, watch_flow_run
from prefect.client import Client
from prefect.engine.signals import signal_from_state
from prefect.run_configs import RunConfig
from prefect.utilities.graphql import EnumValue, with_args
from prefect.utilities.tasks import defaults_from_attrs


@task
def create_flow_run(
    flow_id: str = None,
    flow_name: str = None,
    project_name: str = "",
    parameters: dict = None,
    context: dict = None,
    labels: Iterable[str] = None,
    run_name: str = None,
    run_config: Optional[RunConfig] = None,
    scheduled_start_time: Optional[
        Union[
            pendulum.DateTime,
            datetime.datetime,
            pendulum.Duration,
            datetime.timedelta,
        ]
    ] = None,
    idempotency_key: str = None,
) -> str:
    """
    Task to create a flow run in the Prefect backend.

    The flow to run must be registered and an agent must be available to deploy the
    flow run.

    Args:
        - flow_id: The flow or flow group uuid to lookup the flow to run
        - flow_name: The flow name to lookup the flow to run
        - project_name: The project name to lookup the flow to run. For use with
            `flow_name` if you have flows with the same name in multiple projects
        - parameters: An optional dictionary of parameters to pass to the flow run
        - context: An optional dictionary of context variables to pass to the flow run
        - labels: An optional iterable of labels to set on the flow run; if not
            provided, the default set of labels for the flow will be used
        - run_name: An optional name for the flow run; if not provided, the name will
            be generated as "{current_run_name}-{flow_name}"
        - run_config: An optional run config to use for the flow run; will override any
            existing run config settings
        - scheduled_start_time: An optional time in the future to schedule flow run
            execution for. If not provided, the flow run will be scheduled to start now
        - idempotency_key: a unique idempotency key for scheduling the
            flow run. Duplicate flow runs with the same idempotency key will only create
            a single flow run. This is useful for ensuring that only one run is created
            if this task is retried. If not provided, defaults to the active task run
            id and its map index.

    Returns:
        str: The UUID of the created flow run
    """

    if flow_id and flow_name:
        raise ValueError(
            "Received both `flow_id` and `flow_name`. Only one flow identifier "
            "can be passed."
        )
    if not flow_id and not flow_name:
        raise ValueError(
            "Both `flow_id` and `flow_name` are null. You must pass a flow "
            "identifier"
        )

    logger = prefect.context.logger
    logger.debug("Looking up flow metadata...")

    if flow_id:
        flow = FlowView.from_id(flow_id)

    if flow_name:
        flow = FlowView.from_flow_name(flow_name, project_name=project_name)

    # Generate a 'sub-flow' run name
    if not run_name:
        current_run = prefect.context.get("flow_run_name")
        if current_run:
            run_name = f"{current_run}-{flow.name}"

    # A run name for logging display; robust to 'run_name' being empty
    run_name_dsp = run_name or "<generated-name>"

    logger.info(f"Creating flow run {run_name_dsp!r} for flow {flow.name!r}...")

    if idempotency_key is None:
        # Generate a default key, if the context is missing this data just fall through
        # to `None`
        idempotency_key = prefect.context.get("task_run_id")
        map_index = prefect.context.get("map_index")
        if idempotency_key and map_index is not None:
            idempotency_key += f"-{map_index}"

    if isinstance(scheduled_start_time, (pendulum.Duration, datetime.timedelta)):
        scheduled_start_time = pendulum.now("utc") + scheduled_start_time

    client = Client()
    flow_run_id = client.create_flow_run(
        flow_id=flow.flow_id,
        parameters=parameters,
        context=context,
        labels=labels,
        run_name=run_name,
        run_config=run_config,
        scheduled_start_time=scheduled_start_time,
        idempotency_key=idempotency_key,
    )

    run_url = client.get_cloud_url("flow-run", flow_run_id)
    logger.info(f"Created flow run {run_name_dsp!r}: {run_url}")

    return flow_run_id


# Flow run results ---------------------------------------------------------------------


@task
def get_task_run_result(
    flow_run_id: str, task_slug: str, map_index: int = -1, poll_time: int = 5
) -> Any:
    """
    Task to get the result of a task from a flow run.

    Will wait for the flow run to finish entirely or dynamic task run results will not
    be properly populated.

    Results are loaded from the `Result` location of the task which may not be
    accessible from where this task is executed. You will need to ensure results can
    be accessed.

    Args:
        - flow_run_id: The flow run the task run belongs to
        - task_slug: The 'slug' of the task run you want to get the result of
        - map_index: If the task is mapped, the index you would like to access. By
            default, if given a mapped task, all of the child results will be loaded.
        - poll_time: The amount of time to wait while polling to check if the sub-flow
            has finished

    Returns:
        Any: The return value of the task
    """
    logger = prefect.context.logger

    if not task_slug:
        # Catch this explicitly because the user may user `task.slug` which is often
        # null
        raise ValueError("Required argument `task_slug` is empty")

    task_dsp = repr(task_slug) if map_index == -1 else f"'{task_slug}[{map_index}]'"

    if prefect.context.get("flow_run_id") == flow_run_id:
        # Since we are going to wait for flow run completion, if they pass this flow
        # run id, we will hang forever.
        raise ValueError(
            "Given `flow_run_id` is the same as the currently running flow. The "
            "`get_task_run_result` task cannot be used to retrieve results from the "
            "flow run it belongs to."
        )

    # Get the parent flow run state
    flow_run = FlowRunView.from_flow_run_id(flow_run_id)

    # Wait for the flow run to be finished
    while not flow_run.state.is_finished():
        logger.debug(
            f"Waiting for flow run {flow_run_id} to finish before retreiving result "
            f"for task run {task_dsp}..."
        )
        time.sleep(poll_time)
        flow_run = flow_run.get_latest()

    # Get the task run
    logger.debug("Retrieving task run metadata...")
    task_run = flow_run.get_task_run(task_slug=task_slug, map_index=map_index)

    # Load the result from storage
    logger.debug(
        f"Loading task run result from {type(task_run.state._result).__name__}..."
    )
    return task_run.get_result()


@task
def wait_for_flow_run(
    flow_run_id: str,
    stream_states: bool = True,
    stream_logs: bool = False,
    raise_final_state: bool = False,
    max_duration: timedelta = timedelta(hours=12),
) -> "FlowRunView":
    """
    Task to wait for a flow run to finish executing, streaming state and log information

    Args:
        - flow_run_id: The flow run id to wait for
        - stream_states: Stream information about the flow run state changes
        - stream_logs: Stream flow run logs; if `stream_state` is `False` this will be
            ignored
        - raise_final_state: If set, the state of this task will be set to the final
            state of the child flow run on completion.
        - max_duration: Duration to wait for flow run to complete. Defaults to 12 hours.

    Returns:
        FlowRunView: A view of the flow run after completion
    """

    flow_run = FlowRunView.from_flow_run_id(flow_run_id)

    for log in watch_flow_run(
        flow_run_id,
        stream_states=stream_states,
        stream_logs=stream_logs,
        max_duration=max_duration,
    ):
        message = f"Flow {flow_run.name!r}: {log.message}"
        prefect.context.logger.log(log.level, message)

    # Get the final view of the flow run
    flow_run = flow_run.get_latest()

    if raise_final_state:
        state_signal = signal_from_state(flow_run.state)(
            message=f"{flow_run_id} finished in state {flow_run.state}",
            result=flow_run,
        )
        raise state_signal
    else:
        return flow_run


# Legacy -------------------------------------------------------------------------------


class StartFlowRun(Task):
    """
    Task used to kick off a flow run using Prefect Core's server or Prefect Cloud.  If multiple
    versions of the flow are found, this task will kick off the most recent unarchived version.

    Args:
        - flow_name (str, optional): the name of the flow to schedule; this value may also be
            provided at run time
        - project_name (str, optional): the name of the project in which the flow is located;
            this value may also be provided at runtime.
        - parameters (dict, optional): the parameters to pass to the flow run being scheduled;
            this value may also be provided at run time
        - run_config (RunConfig, optional): a run-config to use for this flow
            run, overriding any existing flow settings.
        - wait (bool, optional): whether to wait the triggered flow run's state; if True, this
            task will wait until the flow run is complete, and then reflect the corresponding
            state as the state of this task.  Defaults to `False`.
        - new_flow_context (dict, optional): the optional run context for the new flow run
        - run_name (str, optional): name to be set for the flow run
        - scheduled_start_time (datetime, optional): the time to schedule the execution
            for; if not provided, defaults to now
        - poll_interval (timedelta): the time to wait between each check if the flow is finished.
                Has to be >= 3 seconds. Used only if `wait=True`. Defaults to 10 seconds.
        - create_link_artifact (bool, optional): whether to create a link artifact
            to the child flow run page.  Defaults to `True`.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        flow_name: str = None,
        project_name: str = None,
        parameters: dict = None,
        run_config: RunConfig = None,
        wait: bool = False,
        new_flow_context: dict = None,
        run_name: str = None,
        scheduled_start_time: datetime.datetime = None,
        poll_interval: timedelta = timedelta(seconds=10),
        create_link_artifact: bool = True,
        **kwargs: Any,
    ):
        self.flow_name = flow_name
        self.project_name = project_name
        # Check that users haven't passed tasks to `parameters`
        if parameters is not None:
            for v in parameters.values():
                if isinstance(v, Task):
                    raise TypeError(
                        "An instance of `Task` was passed to the `StartFlowRun` constructor via the "
                        "`parameters` kwarg. You'll want to pass these parameters when calling the "
                        "task instead. For example:\n\n"
                        "  start_flow_run = StartFlowRun(...)  # static (non-Task) args go here\n"
                        "  res = start_flow_run(parameters=...)  # dynamic (Task) args go here\n\n"
                        "see https://docs.prefect.io/core/concepts/flows.html#apis for more info."
                    )
        self.parameters = parameters
        self.run_config = run_config
        self.new_flow_context = new_flow_context
        self.run_name = run_name
        self.wait = wait
        self.scheduled_start_time = scheduled_start_time
        if poll_interval.total_seconds() < 3:
            raise ValueError(
                "`poll_interval` needs to be at least 3 seconds to avoid spamming the Prefect server. "
                f"(poll_interval == {poll_interval.total_seconds()} seconds)!"
            )
        self.poll_interval = poll_interval
        self.create_link_artifact = create_link_artifact
        if flow_name:
            kwargs.setdefault("name", f"Flow {flow_name}")
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "flow_name",
        "project_name",
        "parameters",
        "run_config",
        "new_flow_context",
        "run_name",
        "scheduled_start_time",
    )
    def run(
        self,
        flow_name: str = None,
        project_name: str = None,
        parameters: dict = None,
        run_config: RunConfig = None,
        new_flow_context: dict = None,
        run_name: str = None,
        idempotency_key: str = None,
        scheduled_start_time: datetime.datetime = None,
    ) -> str:
        """
        Run method for the task; responsible for scheduling the specified flow run.

        Args:
            - flow_name (str, optional): the name of the flow to schedule; if not provided,
                this method will use the flow name provided at initialization
            - project_name (str, optional): the project in which the flow is located; if
                not provided, this method will use the project provided at initialization.
            - parameters (dict, optional): the parameters to pass to the flow run being
                scheduled; if not provided, this method will use the parameters provided at
                initialization
            - run_config (RunConfig, optional): a run-config to use for this flow
                run, overriding any existing flow settings.
            - new_flow_context (dict, optional): the optional run context for the new flow run
            - run_name (str, optional): name to be set for the flow run
            - idempotency_key (str, optional): a unique idempotency key for scheduling the
                flow run. Duplicate flow runs with the same idempotency key will only create
                a single flow run. This is useful for ensuring that only one run is created
                if this task is retried. If not provided, defaults to the active `task_run_id`.
            - scheduled_start_time (datetime, optional): the time to schedule the execution
                for; if not provided, defaults to now

        Returns:
            - str: the ID of the newly-scheduled flow run

        Raises:
            - ValueError: if flow was not provided, cannot be found, or if a project name was
                not provided while using Cloud as a backend

        Example:
            ```python
            from prefect.tasks.prefect.flow_run import StartFlowRun

            kickoff_task = StartFlowRun(project_name="Hello, World!", flow_name="My Cloud Flow")
            ```

        """

        # verify that flow and project names were passed where necessary
        if flow_name is None:
            raise ValueError("Must provide a flow name.")
        if project_name is None:
            raise ValueError("Must provide a project name.")

        where_clause = {
            "name": {"_eq": flow_name},
            "archived": {"_eq": False},
            "project": {"name": {"_eq": project_name}},
        }

        # find the flow ID to schedule
        query = {
            "query": {
                with_args(
                    "flow",
                    {
                        "where": where_clause,
                        "order_by": {"version": EnumValue("desc")},
                        "limit": 1,
                    },
                ): {"id"}
            }
        }

        client = Client()
        flow = client.graphql(query).data.flow

        # verify that a flow has been returned
        if not flow:
            raise ValueError("Flow '{}' not found.".format(flow_name))

        # grab the ID for the most recent version
        flow_id = flow[0].id

        if idempotency_key is None:
            idempotency_key = prefect.context.get("task_run_id", None)

        # providing an idempotency key ensures that retries for this task
        # will not create additional flow runs
        flow_run_id = client.create_flow_run(
            flow_id=flow_id,
            parameters=parameters,
            run_config=run_config,
            idempotency_key=idempotency_key,
            context=new_flow_context,
            run_name=run_name,
            scheduled_start_time=scheduled_start_time,
        )

        self.logger.debug(f"Flow Run {flow_run_id} created.")

        self.logger.debug(f"Creating link artifact for Flow Run {flow_run_id}.")
        run_link = client.get_cloud_url("flow-run", flow_run_id)
        if self.create_link_artifact:
            create_link_artifact(urlparse(run_link).path)
        self.logger.info(f"Flow Run: {run_link}")

        if not self.wait:
            return flow_run_id

        while True:
            time.sleep(self.poll_interval.total_seconds())
            flow_run_state = client.get_flow_run_info(flow_run_id).state
            if flow_run_state.is_finished():
                exc = signal_from_state(flow_run_state)(
                    f"{flow_run_id} finished in state {flow_run_state}"
                )
                raise exc



from time import sleep
import uuid

import requests
from requests import RequestException

import logging
from typing import Union

from prefect import Task
from prefect.engine.signals import FAIL
from prefect.utilities.tasks import defaults_from_attrs


class ConnectionNotFoundException(Exception):
    pass


class AirbyteServerNotHealthyException(Exception):
    pass


class JobNotFoundException(Exception):
    pass


class AirbyteSyncJobFailed(Exception):
    pass


class AirbyteExportConfigurationFailed(Exception):
    pass


class AirbyteClient:
    """
    Esablishes a session with an Airbyte instance and evaluates its current health
    status.

    This client assumes that you're using Airbyte Open-Source, since "For
    Airbyte Open-Source you don't need the API Token for
    Authentication! All endpoints are possible to access using the
    API without it."
    For more information refer to the [Airbyte docs](https://docs.airbyte.io/api-documentation).

    Args:
        - airbyte_base_url (str, mandatory): base api endpoint url for airbyte.
          ex. http://localhost:8000/api/v1

    Returns:
        - session connection with Airbyte
    """

    def __init__(self, logger, airbyte_base_url: str):
        self.airbyte_base_url = airbyte_base_url
        self.logger = logger

    def _establish_session(self):
        session = requests.Session()
        if self._check_health_status(session):
            return session

    def _check_health_status(self, session):
        get_connection_url = self.airbyte_base_url + "/health/"
        try:
            response = session.get(get_connection_url)
            self.logger.debug("Health check response: %s", response.json())
            key = "available" if "available" in response.json() else "db"
            health_status = response.json()[key]
            if not health_status:
                raise AirbyteServerNotHealthyException(
                    f"Airbyte Server health status: {health_status}"
                )
            return True
        except RequestException as e:
            raise AirbyteServerNotHealthyException(e)

    def _export_configuration(self, session) -> bytearray:
        """
        Trigger an export of Airbyte configuration, see:
        https://airbyte-public-api-docs.s3.us-east-2.amazonaws.com/rapidoc-api-docs.html#post-/v1/deployment/export

        Args:
            - session: requests session with which to make call to the Airbyte server.
            - airbyte_base_url: URL of Airbyte server.

        Returns:
            - byte array of Airbyte configuration data
        """
        get_connection_url = self.airbyte_base_url + "/deployment/export/"

        try:
            response = session.post(get_connection_url)
            if response.status_code == 200:
                self.logger.debug("Export configuration response: %s", response)
                export_config = response.content
                return export_config
        except RequestException as e:
            raise AirbyteExportConfigurationFailed(e)


class AirbyteConnectionTask(Task):
    """
    Task for triggering Airbyte Connections, where "A connection is
    a configuration for syncing data between a source and a destination."
    For more information refer to the
    [Airbyte docs](https://docs.airbyte.io/understanding-airbyte/connections)

    This task assumes that the Airbyte Open-Source, since "For
    Airbyte Open-Source you don't need the API Token for
    Authentication! All endpoints are possible to access using the
    API without it."
    For more information refer to the [Airbyte docs](https://docs.airbyte.io/api-documentation)

    Args:
        - airbyte_server_host (str, optional): Hostname of Airbyte server where connection is configured.
            Defaults to localhost.
        - airbyte_server_port (str, optional): Port that the Airbyte server is listening on.
            Defaults to 8000.
        - airbyte_api_version (str, optional): Version of Airbyte API to use to trigger connection sync.
            Defaults to v1.
        - connection_id (str, optional): Default connection id to
            use for sync jobs, if none is specified to `run`.
        - stream_output (Union[bool, int, str], optional): specifies whether this task should log
            the output as it occurs, and at what logging level. If `True` is passed,
            the logging level defaults to `INFO`; otherwise, any integer or string
            value that's passed will be treated as the log level, provided
            the `logging` library can successfully interpret it.
        - **kwargs (Any, optional): additional kwargs to pass to the
            base Task constructor
    """

    # Connection statuses
    CONNECTION_STATUS_ACTIVE = "active"
    CONNECTION_STATUS_INACTIVE = "inactive"
    CONNECTION_STATUS_DEPRECATED = "deprecated"

    # Job statuses
    JOB_STATUS_SUCCEEDED = "succeeded"
    JOB_STATUS_FAILED = "failed"
    JOB_STATUS_PENDING = "pending"

    def __init__(
        self,
        airbyte_server_host: str = "localhost",
        airbyte_server_port: int = 8000,
        airbyte_api_version: str = "v1",
        connection_id: str = None,
        stream_output: Union[bool, int, str] = False,
        **kwargs,
    ):
        self.airbyte_server_host = airbyte_server_host
        self.airbyte_server_port = airbyte_server_port
        self.airbyte_api_version = airbyte_api_version
        self.connection_id = connection_id

        if isinstance(stream_output, str):
            stream_output = logging.getLevelName(stream_output)
            if not isinstance(stream_output, int):
                raise TypeError(
                    f"'stream_output': {stream_output} is not a valid log level"
                )

        self.stream_output = logging.INFO if stream_output is True else stream_output

        super().__init__(**kwargs)

    def _get_connection_status(self, session, airbyte_base_url, connection_id):
        get_connection_url = airbyte_base_url + "/connections/get/"

        # TODO - Missing authentication because Airbyte servers currently do not support authentication
        try:
            response = session.post(
                get_connection_url, json={"connectionId": connection_id}
            )
            self.logger.log(level=self.stream_output, msg=response.json())

            response.raise_for_status()

            # check whether a schedule exists ...
            schedule = response.json().get("schedule")
            if schedule:
                self.logger.warning("Found existing Connection schedule, removing ...")

                # mandatory fields for Connection update ...
                sync_catalog = response.json()["syncCatalog"]
                connection_status = response.json()["status"]

                update_connection_url = airbyte_base_url + "/connections" "/update/"
                response2 = session.post(
                    update_connection_url,
                    json={
                        "connectionId": connection_id,
                        "syncCatalog": sync_catalog,
                        "schedule": None,
                        "status": connection_status,
                    },
                )
                self.logger.log(level=self.stream_output, msg=response2.json())

                if response2.status_code == 200:
                    self.logger.info("Schedule removed ok.")
                else:
                    self.logger.warning("Schedule not removed.")

            connection_status = response.json()["status"]
            return connection_status
        except RequestException as e:
            raise AirbyteServerNotHealthyException(e)

    def _trigger_manual_sync_connection(self, session, airbyte_base_url, connection_id):
        """
        Trigger a manual sync of the Connection, see:
        https://airbyte-public-api-docs.s3.us-east-2.amazonaws.com/rapidoc
        -api-docs.html#post-/v1/connections/sync

        Args:
            session: requests session with which to make call to Airbyte server
            airbyte_base_url: URL of Airbyte server
            connection_id: ID of connection to sync

        Returns: created_at - timestamp of sync job creation

        """
        get_connection_url = airbyte_base_url + "/connections/sync/"

        # TODO - missing authentication ...
        try:
            response = session.post(
                get_connection_url, json={"connectionId": connection_id}
            )
            if response.status_code == 200:
                self.logger.log(level=self.stream_output, msg=response.json())
                job_id = response.json()["job"]["id"]
                job_created_at = response.json()["job"]["createdAt"]
                return job_id, job_created_at
            elif response.status_code == 404:
                # connection_id not found
                self.logger.warning(
                    f"Connection {connection_id} not found, please double "
                    f"check the connection_id ..."
                )
                raise ConnectionNotFoundException(
                    f"Connection {connection_id} not found, please double "
                    f"check the connection_id ..."
                )
        except RequestException as e:
            raise AirbyteServerNotHealthyException(e)

    def _get_job_status(self, session, airbyte_base_url, job_id):
        get_connection_url = airbyte_base_url + "/jobs/get/"
        try:
            response = session.post(get_connection_url, json={"id": job_id})
            if response.status_code == 200:
                self.logger.log(level=self.stream_output, msg=response.json())
                job_status = response.json()["job"]["status"]
                job_created_at = response.json()["job"]["createdAt"]
                job_updated_at = response.json()["job"]["updatedAt"]
                return job_status, job_created_at, job_updated_at
            elif response.status_code == 404:
                self.logger.error(f"Job {job_id} not found...")
                raise JobNotFoundException(f"Job {job_id} not found...")
        except RequestException as e:
            raise AirbyteServerNotHealthyException(e)

    @defaults_from_attrs(
        "airbyte_server_host",
        "airbyte_server_port",
        "airbyte_api_version",
        "connection_id",
    )
    def run(
        self,
        airbyte_server_host: str = None,
        airbyte_server_port: int = None,
        airbyte_api_version: str = None,
        connection_id: str = None,
        poll_interval_s: int = 15,
    ) -> dict:
        """
        Task run method for triggering an Airbyte Connection.

        *It is assumed that the user will have previously configured
        a Source & Destination into a Connection.*
        e.g. MySql -> CSV

        An invocation of `run` will attempt to start a sync job for
        the specified `connection_id` representing the Connection in
        Airbyte.

        `run` will poll Airbyte Server for the Connection status and
        will only complete when the sync has completed or
        when it receives an error status code from an API call.

        Args:
            - airbyte_server_host (str, optional): Hostname of Airbyte server where connection is
                configured. Will overwrite the value provided at init if provided.
            - airbyte_server_port (str, optional): Port that the Airbyte server is listening on.
                Will overwrite the value provided at init if provided.
            - airbyte_api_version (str, optional): Version of Airbyte API to use to trigger connection
                sync. Will overwrite the value provided at init if provided.
            - connection_id (str, optional): if provided,
                will overwrite the value provided at init.
            - poll_interval_s (int, optional): this task polls the
                Airbyte API for status, if provided this value will
                override the default polling time of 15 seconds.

        Returns:
            - dict: connection_id (str) and succeeded_at (timestamp str)
        """
        if not connection_id:
            raise ValueError(
                "Value for parameter `connection_id` *must* \
            be provided."
            )

        try:
            uuid.UUID(connection_id)
        except (TypeError, ValueError):
            raise ValueError(
                "Parameter `connection_id` *must* be a valid UUID \
                i.e. 32 hex characters, including hyphens."
            )

        # see https://airbyte-public-api-docs.s3.us-east-2.amazonaws.com
        # /rapidoc-api-docs.html#overview
        airbyte_base_url = (
            f"http://{airbyte_server_host}:"
            f"{airbyte_server_port}/api/{airbyte_api_version}"
        )

        airbyte = AirbyteClient(self.logger, airbyte_base_url)
        session = airbyte._establish_session()

        self.logger.info(
            f"Getting Airbyte Connection {connection_id}, poll interval "
            f"{poll_interval_s} seconds, airbyte_base_url {airbyte_base_url}"
        )

        connection_status = self._get_connection_status(
            session, airbyte_base_url, connection_id
        )
        if connection_status == self.CONNECTION_STATUS_ACTIVE:
            # Trigger manual sync on the Connection ...
            job_id, job_created_at = self._trigger_manual_sync_connection(
                session, airbyte_base_url, connection_id
            )

            job_status = self.JOB_STATUS_PENDING

            while job_status not in [self.JOB_STATUS_FAILED, self.JOB_STATUS_SUCCEEDED]:
                job_status, job_created_at, job_updated_at = self._get_job_status(
                    session, airbyte_base_url, job_id
                )

                # pending┃running┃incomplete┃failed┃succeeded┃cancelled
                if job_status == self.JOB_STATUS_SUCCEEDED:
                    self.logger.info(f"Job {job_id} succeeded.")
                elif job_status == self.JOB_STATUS_FAILED:
                    self.logger.error(f"Job {job_id} failed.")
                    raise AirbyteSyncJobFailed(f"Job {job_id} failed.")
                else:
                    # wait for next poll interval
                    sleep(poll_interval_s)

            return {
                "connection_id": connection_id,
                "status": connection_status,
                "job_status": job_status,
                "job_created_at": job_created_at,
                "job_updated_at": job_updated_at,
            }
        elif connection_status == self.CONNECTION_STATUS_INACTIVE:
            self.logger.error(
                f"Please enable the Connection {connection_id} in Airbyte Server."
            )
            raise FAIL(
                f"Please enable the Connection {connection_id} in Airbyte Server."
            )
        elif connection_status == self.CONNECTION_STATUS_DEPRECATED:
            self.logger.error(f"Connection {connection_id} is deprecated.")
            raise FAIL(f"Connection {connection_id} is deprecated.")


class AirbyteConfigurationExport(Task):
    """
    Task for triggering an export of the Airbyte configuration.

    This task assumes that you are using Airbyte Open-Source, since "For
    Airbyte Open-Source you don't need the API Token for
    Authentication! All endpoints are possible to access using the
    API without it."
    For more information refer to the [Airbyte docs](https://docs.airbyte.io/api-documentation).

    Args:
        - airbyte_server_host (str, optional): Hostname of Airbyte server where connection is configured.
            Defaults to localhost.
        - airbyte_server_port (str, optional): Port that the Airbyte server is listening on.
            Defaults to 8000.
        - airbyte_api_version (str, optional): Version of Airbyte API to use to trigger connection sync.
            Defaults to v1.
        - **kwargs (Any, optional): Additional kwargs to pass to the
            base Task constructor.

    Returns:
        - compressed gz file byte array of Airbyte configuration data.
        Airbyte requires this file type (or .tar) for configuration imports.

    Example:
        from prefect import Flow
        from prefect.tasks.airbyte import AirbyteConfigurationExport
        import gzip

        @task
        def unzip(export):
            with gzip.open('airbyte.gz', 'wb') as f:
                f.write(export)

        airbyte = AirbyteConfigurationExport()

        with Flow("airbyte_export") as flow:
            export = airbyte()
            unzip(export)
    """

    def __init__(
        self,
        airbyte_server_host: str = "localhost",
        airbyte_server_port: int = 8000,
        airbyte_api_version: str = "v1",
        **kwargs,
    ):
        self.airbyte_server_host = airbyte_server_host
        self.airbyte_server_port = airbyte_server_port
        self.airbyte_api_version = airbyte_api_version
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "airbyte_server_host", "airbyte_server_port", "airbyte_api_version"
    )
    def run(
        self,
        airbyte_server_host: str = None,
        airbyte_server_port: int = None,
        airbyte_api_version: str = None,
    ) -> bytearray:
        """
        Task run method for triggering an export of an Airbyte configuration

        Args:
            - airbyte_server_host (str, optional): Hostname of Airbyte server where connection is
                configured. Will overwrite the value provided at init if provided.
            - airbyte_server_port (str, optional): Port that the Airbyte server is listening on.
                Will overwrite the value provided at init if provided.
            - airbyte_api_version (str, optional): Version of Airbyte API to use to trigger connection
                sync. Will overwrite the value provided at init if provided.

        Returns:
            - byte array of Airbyte configuration data
        """

        airbyte_base_url = (
            f"http://{airbyte_server_host}:"
            f"{airbyte_server_port}/api/{airbyte_api_version}"
        )

        airbyte = AirbyteClient(self.logger, airbyte_base_url)
        session = airbyte._establish_session()

        self.logger.info("Initiating export of Airbyte configuration")
        airbyte_config = airbyte._export_configuration(session)

        return airbyte_config



"""
This module contains a collection of tasks to run Data Quality tests using soda-sql library
"""
from warnings import warn

warn(
    f"soda-sql has been deprecated. We cannot guarantee that the SodaSparkScan "
    "task will continue to function properly.",
    DeprecationWarning,
    stacklevel=2,
)

try:
    from prefect.tasks.sodasql.sodasql_tasks import SodaSQLScan
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.sodasql` requires Prefect to be installed with the "sodasql" extra.'
    ) from err

__all__ = ["SodaSQLScan"]



__all__ = [
    "download_json",
]

import requests
from prefect import task
from typing import Any


@task
def download_json(url: str) -> Any:
    """
    Task to download and parse JSON data stored at a given URL.
    Args:
        - url (str): URL to download.

    Returns:
        - Any: the content at this URL, being parsed from JSON.
    """
    response = requests.get(url)
    response.raise_for_status()
    return response.json()



from sodaspark import scan

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class SodaSparkScan(Task):
    """
    Task for running a SodaSpark scan given a scan definition and a Spark Dataframe.
    For information about SodaSpark please refer to https://docs.soda.io/soda-spark/install-and-use.html.
    SodaSpark uses PySpark under the hood, hence you need Java to be installed on
    the machine where you run this task.

    Args:
        - scan_def (str, optional): scan definition.
          Can be either a path to a YAML file containing the scan definition.
          Please refer to https://docs.soda.io/soda-sql/scan-yaml.html for more information.
          or the scan definition given as a valid YAML string
        - df (pyspark.sql.DataFrame, optional): Spark DataFrame.
          DataFrame where to run tests defined in the scan definition.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(self, scan_def: str = None, df=None, **kwargs):
        self.scan_def = scan_def
        self.df = df
        super().__init__(**kwargs)

    @defaults_from_attrs("scan_def", "df")
    def run(self, scan_def: str = None, df=None):
        """
        Task run method. Execute a scan against a Spark DataFrame.

        Args:
            - scan_def (str, optional): scan definition.
              Can be either a path to a YAML file containing the scan definition.
              Please refer to https://docs.soda.io/soda-sql/scan-yaml.html for more information.
              or the scan definition given as a valid YAML string
            - df (pyspark.sql.DataFrame, optional): Spark DataFrame.
              DataFrame where to run tests defined in the scan definition.

        Returns:
            - A SodaSpark ScanResult
        Raises:
            - ValueError if scan_def is None or not a valid path to a YAML file
            - ValueError if df is None
        """
        if scan_def is None:
            raise ValueError("scan_def cannot be None")
        if df is None:
            raise ValueError("df cannot be None")

        result = scan.execute(scan_definition=scan_def, df=df)
        return result



from typing import List, Dict, Any
from dataclasses import dataclass

import marshmallow

import prefect
from prefect.run_configs.base import RunConfig
from prefect.serialization.flow import FlowSchema
from prefect.serialization.run_config import RunConfigSchema
from prefect.serialization.storage import StorageSchema
from prefect.utilities.graphql import with_args, EnumValue
from prefect.utilities.logging import get_logger


logger = get_logger("backend.flow")


@dataclass(frozen=True)
class FlowView:
    """
    A view of Flow metadata stored in the Prefect API.

    This object is designed to be an immutable view of the data stored in the Prefect
    backend API at the time it is created

    Args:
        - flow_id: The uuid of the flow
        - settings: A dict of flow settings
        - run_config: A dict representation of the flow's run configuration
        - serialized_flow: A serialized copy of the flow
        - archived: A bool indicating if this flow is archived or not
        - project_name: The name of the project the flow is registered to
        - core_version: The core version that was used to register the flow
        - storage: The deserialized Storage object used to store this flow
        - name: The name of the flow
        - flow_group_labels: Labels that are assigned to the parent flow group
    """

    flow_id: str
    settings: dict
    run_config: RunConfig
    serialized_flow: dict
    archived: bool
    project_name: str
    core_version: str
    storage: prefect.storage.Storage
    name: str
    flow_group_labels: List[str]

    @property
    def flow(self) -> "prefect.Flow":
        """
        Deserialize the flow from the backend into a 'Flow' object.

        The deserialized flow is not expected to contain all of the same data as the flow
        object that was originally registered. The backend is free to manipulate the
        serialized flow for optimization.
        """
        # Perform this deserialization lazily (in a property) because it is more likely
        # to fail and we do not want to break the entire view because of bad flow data
        return FlowSchema().load(
            data=self.serialized_flow, partial=True, unknown=marshmallow.EXCLUDE
        )

    @classmethod
    def _from_flow_data(cls, flow_data: dict, **kwargs: Any) -> "FlowView":
        """
        Instantiate a `FlowView` from serialized data

        This method deserializes objects into their Prefect types.

        Args:
            - flow_data: The dict of serialized data
            - **kwargs: Additional kwargs are passed to __init__ and overrides attributes
                from `flow_data`
        """
        flow_data = flow_data.copy()

        flow_id = flow_data.pop("id")
        flow_group_data = flow_data.pop("flow_group")
        flow_group_labels = flow_group_data["labels"]
        project_name = flow_data.pop("project")["name"]
        storage = StorageSchema().load(flow_data.pop("storage"))
        run_config_data = flow_data.pop("run_config")
        run_config = (
            RunConfigSchema().load(run_config_data) if run_config_data else None
        )

        # Combine the data from `flow_data` with `kwargs`
        flow_args = {
            **dict(
                flow_id=flow_id,
                project_name=project_name,
                storage=storage,
                flow_group_labels=flow_group_labels,
                run_config=run_config,
                **flow_data,
            ),
            **kwargs,
        }

        return cls(**flow_args)

    @classmethod
    def from_id(cls, flow_id: str) -> "FlowView":
        """
        Get an instance of this class given a `flow_id` or `flow_group_id` to lookup.
        The `flow_id` will be tried first.

        Args:
            - flow_id: The uuid of the flow or the flow group

        Returns:
            A new instance of FlowView
        """

        try:
            flow = FlowView.from_flow_id(flow_id)
        except ValueError:
            pass
        else:
            return flow

        try:
            flow = FlowView.from_flow_group_id(flow_id)
        except ValueError:
            pass
        else:
            return flow

        raise ValueError(
            f"Given id {flow_id!r} is not an existing flow or flow group id."
        )

    @classmethod
    def from_flow_id(cls, flow_id: str) -> "FlowView":
        """
        Get an instance of this class given a `flow_id` to lookup

        Args:
            - flow_id: The uuid of the flow

        Returns:
            A new instance of FlowView
        """
        if not isinstance(flow_id, str):
            raise TypeError(
                f"Unexpected type {type(flow_id)!r} for `flow_id`, " f"expected 'str'."
            )

        return cls._from_flow_data(cls._query_for_flow(where={"id": {"_eq": flow_id}}))

    @classmethod
    def from_flow_group_id(cls, flow_group_id: str) -> "FlowView":
        """
        Get an instance of this class given a `flow_group_id` to lookup; the newest
        flow in the flow group will be retrieved

        Args:
            - flow_group_id: The uuid of the flow group

        Returns:
            A new instance of FlowView
        """
        if not isinstance(flow_group_id, str):
            raise TypeError(
                f"Unexpected type {type(flow_group_id)!r} for `flow_group_id`, "
                f"expected 'str'."
            )

        return cls._from_flow_data(
            # Get the most recently created flow in the group
            cls._query_for_flows(
                where={"flow_group_id": {"_eq": flow_group_id}},
                order_by={"created": EnumValue("desc")},
            )[0]
        )

    @classmethod
    def from_flow_name(
        cls, flow_name: str, project_name: str = "", last_updated: bool = False
    ) -> "FlowView":
        """
        Get an instance of this class given a flow name. Optionally, a project name can
        be included since flow names are not guaranteed to be unique across projects.

        Args:
            - flow_name: The name of the flow to lookup
            - project_name: The name of the project to lookup. If `None`, flows with an
                explicitly null project will be searched. If `""` (default), the
                lookup will be across all projects.
            - last_updated: By default, if multiple flows are found an error will be
                thrown. If `True`, the most recently updated flow will be returned
                instead.

        Returns:
            A new instance of FlowView
        """
        where: Dict[str, Any] = {"name": {"_eq": flow_name}, "archived": {"_eq": False}}
        if project_name != "":
            where["project"] = {
                "name": ({"_eq": project_name} if project_name else {"_is_null": True})
            }

        flows = cls._query_for_flows(
            where=where,
            order_by={"created": EnumValue("desc")},
        )
        if len(flows) > 1 and not last_updated:
            raise ValueError(
                f"Found multiple flows matching {where}. "
                "Provide a `project_name` as well or toggle `last_updated` "
                "to use the flow that was most recently updated"
            )

        flow = flows[0]
        return cls._from_flow_data(flow)

    @staticmethod
    def _query_for_flow(where: dict, **kwargs: Any) -> dict:
        """
        Query for flow data using `_query_for_flows` but throw an exception if
        more than one matching flow is found

        Args:
            - where: The `where` clause to use
            - **kwargs: Additional kwargs are passed to `_query_for_flows`

        Returns:
            A dict of flow data
        """
        flows = FlowView._query_for_flows(where=where, **kwargs)

        if len(flows) > 1:
            raise ValueError(
                f"Found multiple ({len(flows)}) flows while querying for flows "
                f"where {where}: {flows}"
            )

        if not flows:
            return {}

        flow = flows[0]
        return flow

    @staticmethod
    def _query_for_flows(
        where: dict,
        order_by: dict = None,
        error_on_empty: bool = True,
    ) -> List[dict]:
        """
        Query for flow data necessary to initialize `Flow` instances with
        `_Flow.from_flow_data`.

        Args:
            - where (required): The Hasura `where` clause to filter by
            - order_by (optional): An optional Hasura `order_by` clause to order
                 results by
            - error_on_empty (optional): If `True` and no flows are found, a
                `ValueError` will be raised

        Returns:
            A list of dicts of flow information
        """
        client = prefect.Client()

        query_args = {"where": where}
        if order_by is not None:
            query_args["order_by"] = order_by

        flow_query = {
            "query": {
                with_args("flow", query_args): {
                    "id": True,
                    "settings": True,
                    "run_config": True,
                    "serialized_flow": True,
                    "name": True,
                    "archived": True,
                    "project": {"name"},
                    "core_version": True,
                    "storage": True,
                    "flow_group": {"labels"},
                }
            }
        }

        result = client.graphql(flow_query)
        flows = result.get("data", {}).get("flow", None)

        if flows is None:
            raise ValueError(
                f"Received bad result while querying for flows where {where}: "
                f"{result}"
            )

        if not flows:  # Empty list
            if error_on_empty:
                raise ValueError(
                    f"No results found while querying for flows where {where!r}"
                )
            return []

        # Return a list
        return flows

    def __repr__(self) -> str:
        # Implement a shorter repr than dataclass would give us
        return (
            f"{type(self).__name__}"
            "("
            + ", ".join(
                [
                    f"flow_id={self.flow_id!r}",
                    f"name={self.name!r}",
                    f"project_name={self.project_name!r}",
                    f"storage_type={type(self.storage).__name__}",
                ]
            )
            + ")"
        )



"""
Utility functions for interacting with AWS.
"""
import threading
import warnings
import weakref

import prefect

import boto3
from typing import TYPE_CHECKING, Any, Optional, MutableMapping

if TYPE_CHECKING:
    import botocore.client


_CLIENT_CACHE = (
    weakref.WeakValueDictionary()
)  # type: MutableMapping[tuple, botocore.client.BaseClient]
_LOCK = threading.Lock()


def get_boto_client(
    resource: str,
    credentials: Optional[dict] = None,
    region_name: Optional[str] = None,
    profile_name: Optional[str] = None,
    **kwargs: Any
) -> "botocore.client.BaseClient":
    """
    Utility function for loading boto3 client objects from a given set of credentials.

    Note: this utility is threadsafe, and will cache _active_ clients to be
    potentially reused by other concurrent callers, reducing the overhead of
    client creation.

    Args:
        - resource (str): the name of the resource to retrieve a client for
        - credentials (dict, optional): a dictionary of AWS credentials used to
            initialize the Client; if not provided, will attempt to load the
            Client using ambient environment settings
        - region_name (str, optional): The aws region name to use, defaults to your
            global configured default.
        - profile_name (str, optional): The name of a boto3 profile to use.
        - **kwargs (Any, optional): additional keyword arguments to pass to boto3

    Returns:
        - Client: an initialized and authenticated boto3 Client
    """

    if kwargs.pop("use_session", None) is not None:
        # Deprecated in 0.14.9
        warnings.warn(
            "The `use_session` kwarg has been deprecated, prefect now infers "
            "whether a new boto session should be used automatically."
        )

    botocore_session = kwargs.pop("botocore_session", None)

    if credentials:
        aws_access_key = credentials["ACCESS_KEY"]
        aws_secret_access_key = credentials["SECRET_ACCESS_KEY"]
        aws_session_token = credentials.get("SESSION_TOKEN")
    else:
        ctx_credentials = prefect.context.get("secrets", {}).get("AWS_CREDENTIALS", {})
        aws_access_key = ctx_credentials.get("ACCESS_KEY")
        aws_secret_access_key = ctx_credentials.get("SECRET_ACCESS_KEY")
        aws_session_token = ctx_credentials.get("SESSION_TOKEN")

    kwargs_access_key_id = kwargs.pop("aws_access_key_id", None)
    kwargs_secret_access_key = kwargs.pop("aws_secret_access_key", None)
    kwargs_session_token = kwargs.pop("aws_session_token", None)

    aws_access_key = aws_access_key or kwargs_access_key_id
    aws_secret_access_key = aws_secret_access_key or kwargs_secret_access_key
    aws_session_token = aws_session_token or kwargs_session_token

    # Boto3 clients _are_ threadsafe, but the creation of a client or a session
    # isn't thread safe. Creating a client has a small cost (~5-10 ms), while a
    # session has a larger cost (50-120 ms). Creating a new session also
    # renegotiates auth - apparently AWS can throttle these requests if there
    # are too many of them, so we really want to minimize session creation if
    # we can.
    #
    # We keep a weakref cache of clients around to minimize recreation, and use
    # a global lock to synchronize access to this cache. Unfortunately a client
    # doesn't keep a ref around to its backing session, so we can't (easily)
    # keep a cache of sessions around. Most users should use only a single
    # session per process though, and boto3 already supports that with a global
    # shared session. In practice this is much more efficient than creating a
    # new AWS session for every request, and still maintains thread safety.
    with _LOCK:
        botocore_session = kwargs.pop("botocore_session", None)
        cache_key = (
            profile_name,
            region_name,
            id(botocore_session),
            resource,
            aws_access_key,
            aws_secret_access_key,
            aws_session_token,
        )

        # If no extra kwargs and client already created, use the cached client
        if not kwargs and cache_key in _CLIENT_CACHE:
            return _CLIENT_CACHE[cache_key]

        if profile_name or botocore_session:
            session = boto3.session.Session(
                profile_name=profile_name,
                region_name=region_name,
                botocore_session=botocore_session,
            )
        else:
            session = boto3

        client = session.client(
            resource,
            aws_access_key_id=aws_access_key,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
            **kwargs,
        )
        # Cache the client if no extra kwargs
        if not kwargs:
            _CLIENT_CACHE[cache_key] = client

    return client



"""
A collection of tasks for interacting with GitHub.
"""

from .issues import OpenGitHubIssue
from .prs import CreateGitHubPR
from .repos import GetRepoInfo, CreateBranch
from .comments import CreateIssueComment

__all__ = [
    "CreateBranch",
    "CreateGitHubPR",
    "CreateIssueComment",
    "GetRepoInfo",
    "OpenGitHubIssue",
]



import jwt
from requests import Session
import time
from typing import Dict, Union

from prefect.engine.signals import FAIL


class CubeJSClient:

    # Cube Cloud base URL
    __CUBEJS_CLOUD_BASE_URL = "https://{subdomain}.cubecloud.dev"

    def __init__(
        self,
        subdomain: str,
        url: str,
        security_context: Union[str, Dict],
        secret: str,
        wait_api_call_secs: int,
        max_wait_time: int,
    ):
        """
        Initialize a `CubeJSClient`.
        The client can be used to interact with Cube.js APIs.

        Args:
            - subdomain (str): Cube Cloud subdomain.
            - url (str): Cube.js URL (likely to be used in self-hosted Cube.js
                deployments).
            - security_context (str, dict): The security context to be used
                when interacting with Cube.js APIs.
            - secret (str): The secret string to be used, together with the
                `security_context`, to generate the API token to pass in the
                authorization header.
            - wait_api_call_secs (int): Number of seconds to wait
                between API calls.
            - max_wait_time (int): The maximum amount of seconds to wait for
                an API call to respond.
        """
        self.subdomain = subdomain
        self.url = url
        self.security_context = security_context
        self.secret = secret
        self.cube_base_url = self._get_cube_base_url()
        self.api_token = self.get_api_token()
        self.query_api_url = self._get_query_api_url()
        self.generated_sql_api_url = self._get_generated_sql_api_url()
        self.wait_api_call_secs = wait_api_call_secs
        self.max_wait_time = max_wait_time

    def _get_cube_base_url(self) -> str:
        """
        Get Cube.js base URL.

        Returns:
            - Cube.js API base url.
        """
        cube_base_url = self.__CUBEJS_CLOUD_BASE_URL
        if self.subdomain:
            cube_base_url = (
                f"{cube_base_url.format(subdomain=self.subdomain)}/cubejs-api"
            )
        else:
            cube_base_url = self.url
        return cube_base_url

    def _get_query_api_url(self) -> str:
        """
        Get Cube.js Query API URL.

        Returns:
            - Cube.js Query API URL.
        """
        return f"{self.cube_base_url}/v1/load"

    def _get_generated_sql_api_url(self) -> str:
        """
        Get Cube.js Query SQL API URL.

        Returns:
            - Cube.js Query SQL API URL.
        """

        return f"{self.cube_base_url}/v1/sql"

    def get_api_token(self) -> str:
        """
        Build API Token given the security context and the secret.

        Returns:
            - The API Token to include in the authorization headers
                when calling Cube.js APIs.
        """
        api_token = jwt.encode(payload={}, key=self.secret)
        if self.security_context:

            extended_context = self.security_context
            if (
                "exp" not in self.security_context
                and "expiresIn" not in self.security_context
            ):
                extended_context["expiresIn"] = "7d"
            api_token = jwt.encode(
                payload=extended_context, key=self.secret, algorithm="HS256"
            )

        return api_token

    def _get_data_from_url(self, api_url: str, params: Dict) -> Dict:
        """
        Retrieve data from a Cube.js API.

        Args:
            - api_url (str): The URL of the Cube API to call.
            - params (dict): Parameters to be passed to the API call.

        Raises:
            - `prefect.engine.signals.FAIL` if the response has `status_code != 200`.
            - `prefect.engine.signals.FAIL` if the REST APIs takes too long to respond,
                with regards to `max_wait_time`.

        Returns:
            - Cube.js REST API JSON response
        """
        session = Session()
        session.headers = {
            "Content-type": "application/json",
            "Authorization": self.api_token,
        }
        elapsed_wait_time = 0
        while not self.max_wait_time or elapsed_wait_time <= self.max_wait_time:

            with session.get(url=api_url, params=params) as response:
                if response.status_code == 200:
                    data = response.json()

                    if "error" in data.keys() and "Continue wait" in data["error"]:
                        time.sleep(self.wait_api_call_secs)
                        elapsed_wait_time += self.wait_api_call_secs
                        continue

                    else:
                        return data

                else:
                    raise FAIL(
                        message=f"Cube.js load API failed! Error is: {response.reason}"
                    )
        msg = f"Cube.js load API took longer than {self.max_wait_time} seconds to provide a response."
        raise FAIL(message=msg)

    def get_data(
        self,
        params: Dict,
        include_generated_sql: bool,
    ) -> Dict:
        """
        Retrieve data from Cube.js `/load` REST API.

        Args:
            - params (dict): Parameters to pass to the `/load` REST API.
            - include_generated_sql (bool): Whether to include the
                corresponding generated SQL or not.

        Returns:
            - Cube.js `/load` API JSON response, augmented with SQL
                information if `include_generated_sql` is `True`.
        """
        data = self._get_data_from_url(api_url=self.query_api_url, params=params)

        if include_generated_sql:
            data["sql"] = self._get_data_from_url(
                api_url=self.generated_sql_api_url, params=params
            )["sql"]

        return data



import os

from typing import TYPE_CHECKING, Any

from prefect.client import Secret
from prefect.storage import Storage
from prefect.utilities.storage import extract_flow_from_file
from prefect.utilities.git import TemporaryGitRepo

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class Git(Storage):
    """
    Git storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a git repository.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently.

    A typical workflow using this storage type might look like the following:

    - Compose flow `.py` file where flow has Git storage:

    ```python
    flow = Flow("my-flow")
    flow.storage = Git(repo="my/repo", flow_path="/flows/flow.py", repo_host="github.com")
    ```

    - Push this `flow.py` file to the `my/repo` repository under `/flows/flow.py`.

    - Call `prefect register -f flow.py` to register this flow with Git storage.

    Args:
        - flow_path (str): A file path pointing to a .py file containing a flow
        - repo (str, optional): The name of a git repository to store this Flow.
            If not provided, the repo must be set using a secret. See `git_clone_url_secret_name`.
        - repo_host (str, optional): The site hosting the repo. Defaults to 'github.com'
        - flow_name (str, optional): A specific name of a flow to extract from a file.
            If not set then the first flow object retrieved from file will be returned.
        - git_token_secret_name (str, optional): The name of the Prefect Secret containing
            an access token for the repo. Defaults to None
        - git_token_username (str, optional): the username associated with git access token,
            if not provided it will default to repo owner
        - git_clone_url_secret_name (str, optional): the name of the Prefect Secret specifying
            the exact git url to clone, if provided it will override repo, repo_host,
            git_token_secret_name, `git_token_username`, `use_ssh`, and `format_access_token`
            parameters
        - branch_name (str, optional): branch name, if not specified and `tag` and `commit_sha`
            not specified, repo default branch latest commit will be used
        - tag (str, optional): tag name, if not specified and `branch_name` and `commit_sha`
            not specified, repo default branch latest commit will be used
        - commit (str, optional): a commit SHA-1 value, if not specified and `branch_name`
            and `tag` not specified, repo default branch latest commit will be used
        - clone_depth (int): the number of history revisions in cloning, defaults to 1
        - use_ssh (bool): if True, cloning will use ssh. Ssh keys must be correctly
            configured in the environment for this to work
        - format_access_token (bool): if True, the class will attempt to format access tokens
            for common git hosting sites
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        flow_path: str,
        repo: str = None,
        repo_host: str = "github.com",
        flow_name: str = None,
        git_token_secret_name: str = None,
        git_token_username: str = None,
        git_clone_url_secret_name: str = None,
        branch_name: str = None,
        tag: str = None,
        commit: str = None,
        clone_depth: int = 1,
        use_ssh: bool = False,
        format_access_token: bool = True,
        **kwargs: Any,
    ) -> None:
        if sum([bool(x) for x in (branch_name, tag, commit)]) > 1:
            raise ValueError(
                "Please provide only one of the following parameters: `branch_name`, `tag`, `commit`"
            )

        if repo is None and git_clone_url_secret_name is None:
            raise ValueError(
                "Either `repo` or `git_clone_url_secret_name` must be provided"
            )

        if use_ssh and git_token_secret_name is not None:
            self.logger.warning(
                "Git Storage initialized with `use_ssh = True` and `git_token_secret_name` provided. "
                "SSH will be used to clone the repository. `git_token_secret_name` will be ignored"
            )

        if git_clone_url_secret_name and any(
            [
                repo,
                repo_host,
                git_token_secret_name,
                git_token_username,
                use_ssh,
                format_access_token,
            ]
        ):
            self.logger.warning(
                "Git storage initialized with a `git_clone_url_secret_name`. The value of this Secret "
                "will be used to clone the repository, ignoring `repo`, `repo_host`, "
                "`git_token_secret_name`,  `git_token_username`, `use_ssh`, and "
                "`format_access_token`."
            )

        self.flow_path = flow_path
        self.repo = repo
        self.repo_host = repo_host
        self.flow_name = flow_name
        self.git_token_secret_name = git_token_secret_name
        self.git_clone_url_secret_name = git_clone_url_secret_name

        # if not provided, assume the username associated with the token
        # is the organization that owns the repo
        self.git_token_username = (
            git_token_username
            if git_token_username
            else (repo.split("/")[0] if repo else None)
        )

        self.branch_name = branch_name
        self.tag = tag
        self.commit = commit
        self.clone_depth = clone_depth
        self.use_ssh = use_ssh
        self.format_access_token = format_access_token
        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.
        Args:
            - flow_name (str): the name of the flow to return.
        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")

        with TemporaryGitRepo(
            git_clone_url=self.git_clone_url,
            branch_name=self.branch_name,
            tag=self.tag,
            commit=self.commit,
            clone_depth=self.clone_depth,
        ) as temp_repo:
            flow = extract_flow_from_file(
                file_path=os.path.join(temp_repo.temp_dir.name, self.flow_path),
                flow_name=self.flow_name,
            )
        return flow

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.
        Args:
            - flow (Flow): a Prefect Flow to add
        Returns:
            - str: the location of the added flow in the repo
        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )
        self.flows[flow.name] = self.flow_path
        self._flows[flow.name] = flow
        return self.flow_path

    @property
    def git_token_secret(self) -> str:
        """
        Get and format the git secret token if it exists
        """
        if self.git_token_secret_name is None:
            return ""

        # get the access token and format it for common git hosts
        access_token = Secret(self.git_token_secret_name).get()
        if self.format_access_token:
            return f"{self.git_token_username}:{access_token}"
        return str(access_token)

    @property
    def git_clone_url(self) -> str:
        """
        Build the git url to clone
        """
        if self.git_clone_url_secret_name:
            return Secret(self.git_clone_url_secret_name).get()  # type: ignore

        if self.use_ssh:
            return f"git@{self.repo_host}:{self.repo}.git"
        return f"https://{self.git_token_secret}@{self.repo_host}/{self.repo}.git"



"""
This module contains a task for triggering [Airbyte](https://airbyte.io/) connection sync jobs as well as a configuration export
"""
from .airbyte import AirbyteConnectionTask, AirbyteConfigurationExport

__all__ = ["AirbyteConnectionTask", "AirbyteConfigurationExport"]



import sqlite3 as sql
from contextlib import closing
from typing import Any, cast

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class SQLiteQuery(Task):
    """
    Task for executing a single query against a sqlite3 database; returns
    the result (if any) from the query.

    Args:
        - db (str, optional): the location of the database (.db) file
        - query (str, optional): the optional _default_ query to execute at runtime;
            can also be provided as a keyword to `run`, which takes precedence over this default.
            Note that a query should consist of a _single SQL statement_.
        - data (tuple, optional): values to use when `query` is a parametrized string. See
            https://docs.python.org/3/library/sqlite3.html for more details.
        - **kwargs (optional): additional keyword arguments to pass to the
            standard Task initalization
    """

    def __init__(
        self, db: str = None, query: str = None, data: tuple = (), **kwargs: Any
    ):
        self.db = db
        self.query = query
        self.data = data
        super().__init__(**kwargs)

    @defaults_from_attrs("db", "query", "data")
    def run(self, db: str = None, query: str = None, data: tuple = ()):
        """
        Args:
            - db (str, optional): the location of the database (.db) file;
                if not provided, `self.db` will be used instead.
            - query (str, optional): the optional query to execute at runtime;
                if not provided, `self.query` will be used instead. Note that a
                query should consist of a _single SQL statement_.
            - data (tuple, optional): values to use when query is a parametrized string. See
                https://docs.python.org/3/library/sqlite3.html for more details.

        Returns:
            - [Any]: the results of the query
        """
        db = cast(str, db)
        query = cast(str, query)
        data = cast(tuple, data)

        if db is None:
            raise ValueError("The db must be specified")

        if query is None:
            raise ValueError("The query string must be specified")

        if "?" in query and not data:
            raise ValueError(
                "The data inputs must be specified as a tuple if query "
                "contains parametrized values '?'"
            )

        with closing(sql.connect(db)) as conn:
            with closing(conn.cursor()) as cursor:
                cursor.execute(query, data)
                out = cursor.fetchall()
        return out


class SQLiteScript(Task):
    """
    Task for executing a SQL script against a sqlite3 database; does not
    return any values, but useful for creating tables and inserting values into the tables.

    Args:
        - db (str, optional): the location of the database (.db) file
        - script (str, optional): the optional _default_ script string to render at runtime;
            can also be provided as a keyword to `run`, which takes precedence over this default.
        - **kwargs (optional): additional keyword arguments to pass to the
            standard Task initialization
    """

    def __init__(self, db: str = None, script: str = None, **kwargs: Any):
        self.db = db
        self.script = script
        super().__init__(**kwargs)

    @defaults_from_attrs("db", "script")
    def run(self, db: str = None, script: str = None):
        """
        Args:
            - db (str, optional): the location of the database (.db) file;
                if not provided, `self.db` will be used instead.
            - script (str, optional): the optional script to execute at runtime;
                if not provided, `self.script` will be used instead.
        """
        if db is None:
            raise ValueError("The db must be specified")

        if script is None:
            raise ValueError("The query script string must be specified")

        with closing(sql.connect(db)) as conn:
            with closing(conn.cursor()) as cursor:
                cursor.executescript(script)



from pathlib import Path
from typing import List, Union, Optional

from google.cloud import bigquery
from google.cloud.exceptions import NotFound

from prefect.core import Task
from prefect.engine.signals import FAIL, SUCCESS
from prefect.utilities.gcp import get_bigquery_client
from prefect.utilities.tasks import defaults_from_attrs


class BigQueryTask(Task):
    """
    Task for executing queries against a Google BigQuery table and (optionally) returning
    the results.  Note that _all_ initialization settings can be provided / overwritten at runtime.

    Args:
        - query (str, optional): a string of the query to execute
        - query_params (list[tuple], optional): a list of 3-tuples specifying BigQuery query
            parameters; currently only scalar query parameters are supported.  See [the Google
            documentation](https://cloud.google.com/bigquery/docs/parameterized-queries#bigquery-query-params-python)
            for more details on how both the query and the query parameters should be formatted
        - project (str, optional): the project to initialize the BigQuery Client with; if not
            provided, will default to the one inferred from your credentials
        - location (str, optional): location of the dataset that will be queried; defaults to "US"
        - dry_run_max_bytes (int, optional): if provided, the maximum number of bytes the query
            is allowed to process; this will be determined by executing a dry run and raising a
            `ValueError` if the maximum is exceeded
        - dataset_dest (str, optional): the optional name of a destination dataset to write the
            query results to, if you don't want them returned; if provided, `table_dest` must
            also be provided
        - table_dest (str, optional): the optional name of a destination table to write the
            query results to, if you don't want them returned; if provided, `dataset_dest` must also be
            provided
        - to_dataframe (bool, optional): if provided, returns the results of the query as a pandas
            dataframe instead of a list of `bigquery.table.Row` objects. Defaults to False
        - job_config (dict, optional): an optional dictionary of job configuration parameters; note that
            the parameters provided here must be pickleable (e.g., dataset references will be rejected)
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(
        self,
        query: str = None,
        query_params: Optional[List[tuple]] = None,  # 3-tuples
        project: Optional[str] = None,
        location: str = "US",
        dry_run_max_bytes: Optional[int] = None,
        dataset_dest: Optional[str] = None,
        table_dest: Optional[str] = None,
        to_dataframe: Optional[bool] = False,
        job_config: Optional[dict] = None,
        **kwargs,
    ):
        self.query = query
        self.query_params = query_params
        self.project = project
        self.location = location
        self.dry_run_max_bytes = dry_run_max_bytes
        self.dataset_dest = dataset_dest
        self.table_dest = table_dest
        self.to_dataframe = to_dataframe
        self.job_config = job_config or {}
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "query",
        "query_params",
        "project",
        "location",
        "dry_run_max_bytes",
        "dataset_dest",
        "table_dest",
        "to_dataframe",
        "job_config",
    )
    def run(
        self,
        query: str = None,
        query_params: Optional[List[tuple]] = None,
        project: Optional[str] = None,
        location: str = "US",
        dry_run_max_bytes: Optional[int] = None,
        credentials: Optional[dict] = None,
        dataset_dest: Optional[str] = None,
        table_dest: Optional[str] = None,
        to_dataframe: Optional[bool] = False,
        job_config: Optional[dict] = None,
    ):
        """
        Run method for this Task.  Invoked by _calling_ this Task within a Flow context, after
        initialization.

        Args:
            - query (str, optional): a string of the query to execute
            - query_params (list[tuple], optional): a list of 3-tuples specifying BigQuery
                query parameters; currently only scalar query parameters are supported. See
                [the Google
                documentation](https://cloud.google.com/bigquery/docs/parameterized-queries#bigquery-query-params-python)
                for more details on how both the query and the query parameters should be
                formatted
            - project (str, optional): the project to initialize the BigQuery Client with; if
                not provided, will default to the one inferred from your credentials
            - location (str, optional): location of the dataset that will be queried; defaults
                to "US"
            - dry_run_max_bytes (int, optional): if provided, the maximum number of bytes the
                query is allowed to process; this will be determined by executing a dry run and
                raising a `ValueError` if the maximum is exceeded
            - credentials (dict, optional): a JSON document containing Google Cloud credentials.
                You should provide these at runtime with an upstream Secret task.  If not
                provided, Prefect will first check `context` for `GCP_CREDENTIALS` and lastly
                will use default Google client logic.
            - dataset_dest (str, optional): the optional name of a destination dataset to write the
                query results to, if you don't want them returned; if provided, `table_dest`
                must also be provided
            - table_dest (str, optional): the optional name of a destination table to write the
                query results to, if you don't want them returned; if provided, `dataset_dest` must also
                be provided
            - to_dataframe (bool, optional): if provided, returns the results of the query as a pandas
                dataframe instead of a list of `bigquery.table.Row` objects. Defaults to False
            - job_config (dict, optional): an optional dictionary of job configuration parameters; note
                that the parameters provided here must be pickleable (e.g., dataset references will be
                rejected)

        Raises:
            - ValueError: if the `query` is `None`
            - ValueError: if only one of `dataset_dest` / `table_dest` is provided
            - ValueError: if the query will execeed `dry_run_max_bytes`

        Returns:
            - list: a fully populated list of Query results, with one item per row
        """
        # check for any argument inconsistencies
        if query is None:
            raise ValueError("No query provided.")
        if sum([dataset_dest is None, table_dest is None]) == 1:
            raise ValueError(
                "Both `dataset_dest` and `table_dest` must be provided if writing to a "
                "destination table."
            )

        # create client
        client = get_bigquery_client(project=project, credentials=credentials)

        # setup jobconfig
        job_config = bigquery.QueryJobConfig(**job_config)
        if query_params is not None:
            hydrated_params = [
                bigquery.ScalarQueryParameter(*qp) for qp in query_params
            ]
            job_config.query_parameters = hydrated_params

        # perform dry_run if requested
        if dry_run_max_bytes is not None:
            old_info = dict(
                dry_run=job_config.dry_run, use_query_cache=job_config.use_query_cache
            )
            job_config.dry_run = True
            job_config.use_query_cache = False
            self.logger.debug("Performing a dry run...")
            query_job = client.query(query, location=location, job_config=job_config)
            if query_job.total_bytes_processed > dry_run_max_bytes:
                msg = (
                    "Query will process {0} bytes which is above the set maximum of {1} "
                    "for this task."
                ).format(query_job.total_bytes_processed, dry_run_max_bytes)
                raise ValueError(msg)
            job_config.dry_run = old_info["dry_run"]
            job_config.use_query_cache = old_info["use_query_cache"]

        # if writing to a destination table
        if dataset_dest is not None:
            table_ref = client.dataset(dataset_dest).table(table_dest)
            job_config.destination = table_ref

        query_job = client.query(query, location=location, job_config=job_config)

        # if returning the results as a dataframe
        if to_dataframe:
            return query_job.result().to_dataframe()
        # else if returning as a list of bigquery.table.Row objects (default)
        else:
            return list(query_job.result())


class BigQueryStreamingInsert(Task):
    """
    Task for insert records in a Google BigQuery table via [the streaming
    API](https://cloud.google.com/bigquery/streaming-data-into-bigquery).  Note that all of
    these settings can optionally be provided or overwritten at runtime.

    Args:
        - dataset_id (str, optional): the id of a destination dataset to write the
            records to
        - table (str, optional): the name of a destination table to write the
            records to
        - project (str, optional): the project to initialize the BigQuery Client with; if not
            provided, will default to the one inferred from your credentials
        - location (str, optional): location of the dataset that will be written to; defaults
            to "US"
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(
        self,
        dataset_id: str = None,
        table: str = None,
        project: Optional[str] = None,
        location: str = "US",
        **kwargs,
    ):
        self.dataset_id = dataset_id
        self.table = table
        self.project = project
        self.location = location
        super().__init__(**kwargs)

    @defaults_from_attrs("dataset_id", "table", "project", "location")
    def run(
        self,
        records: List[dict],
        dataset_id: str = None,
        table: str = None,
        project: Optional[str] = None,
        location: Optional[str] = "US",
        credentials: Optional[dict] = None,
        **kwargs,
    ):
        """
        Run method for this Task.  Invoked by _calling_ this Task within a Flow context, after
        initialization.

        Args:
            - records (list[dict]): the list of records to insert as rows into the BigQuery
                table; each item in the list should be a dictionary whose keys correspond to
                columns in the table
            - dataset_id (str, optional): the id of a destination dataset to write the records
                to; if not provided here, will default to the one provided at initialization
            - table (str, optional): the name of a destination table to write the
                records to; if not provided here, will default to the one provided at initialization
            - project (str, optional): the project to initialize the BigQuery Client with; if
                not provided, will default to the one inferred from your credentials
            - location (str, optional): location of the dataset that will be written to;
                defaults to "US"
            - credentials (dict, optional): a JSON document containing Google Cloud
                credentials.  You should provide these at runtime with an upstream Secret task.
                If not provided, Prefect will first check `context` for `GCP_CREDENTIALS` and
                lastly will use default Google client logic.
            - **kwargs (optional): additional kwargs to pass to the `insert_rows_json` method;
                see the documentation here:
                https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.client.Client.html

        Raises:
            - ValueError: if all required arguments haven't been provided
            - ValueError: if any of the records result in errors

        Returns:
            - the response from `insert_rows_json`
        """
        # check for any argument inconsistencies
        if dataset_id is None or table is None:
            raise ValueError("Both dataset_id and table must be provided.")

        # create client
        client = get_bigquery_client(project=project, credentials=credentials)

        # get table reference
        table_ref = client.dataset(dataset_id).table(table)

        # stream data in
        response = client.insert_rows_json(table=table_ref, json_rows=records, **kwargs)

        errors = []
        output = []
        for row in response:
            output.append(row)
            if "errors" in row:
                errors.append(row["errors"])

        if errors:
            raise ValueError(errors)

        return output


class BigQueryLoadGoogleCloudStorage(Task):
    """
    Task for insert records in a Google BigQuery table via a [load
    job](https://cloud.google.com/bigquery/docs/loading-data).  Note that all of these settings
    can optionally be provided or overwritten at runtime.

    Args:
        - uri (str, optional): GCS path to load data from
        - dataset_id (str, optional): the id of a destination dataset to write the
            records to
        - table (str, optional): the name of a destination table to write the
            records to
        - project (str, optional): the project to initialize the BigQuery Client with; if not
            provided, will default to the one inferred from your credentials
        - schema (List[bigquery.SchemaField], optional): the schema to use when creating the table
        - location (str, optional): location of the dataset that will be queried; defaults to "US"
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(
        self,
        uri: str = None,
        dataset_id: str = None,
        table: str = None,
        project: Optional[str] = None,
        schema: Optional[List[bigquery.SchemaField]] = None,
        location: str = "US",
        **kwargs,
    ):
        self.uri = uri
        self.dataset_id = dataset_id
        self.table = table
        self.project = project
        self.schema = schema
        self.location = location
        super().__init__(**kwargs)

    @defaults_from_attrs("uri", "dataset_id", "table", "project", "location")
    def run(
        self,
        uri: str = None,
        dataset_id: str = None,
        table: str = None,
        project: Optional[str] = None,
        schema: Optional[List[bigquery.SchemaField]] = None,
        location: str = "US",
        credentials: Optional[dict] = None,
        **kwargs,
    ):
        """
        Run method for this Task.  Invoked by _calling_ this Task within a Flow context, after
        initialization.

        Args:
            - uri (str, optional): GCS path to load data from
            - dataset_id (str, optional): the id of a destination dataset to write the
                records to; if not provided here, will default to the one provided at initialization
            - table (str, optional): the name of a destination table to write the
                records to; if not provided here, will default to the one provided at initialization
            - project (str, optional): the project to initialize the BigQuery Client with; if
                not provided, will default to the one inferred from your credentials
            - schema (List[bigquery.SchemaField], optional): the schema to use when creating
                the table
            - location (str, optional): location of the dataset that will be written to;
                defaults to "US"
            - credentials (dict, optional): a JSON document containing Google Cloud
                credentials.  You should provide these at runtime with an upstream Secret task.
                If not provided, Prefect will first check `context` for `GCP_CREDENTIALS` and
                lastly will use default Google client logic.
            - **kwargs (optional): additional kwargs to pass to the `bigquery.LoadJobConfig`;
                see the documentation here:
                https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.client.Client.html

        Raises:
            - ValueError: if all required arguments haven't been provided
            - ValueError: if the load job results in an error

        Returns:
            - google.cloud.bigquery.job.LoadJob: the response from `load_table_from_uri`
        """
        # check for any argument inconsistencies
        if dataset_id is None or table is None:
            raise ValueError("Both dataset_id and table must be provided.")

        # create client
        client = get_bigquery_client(project=project, credentials=credentials)

        # get table reference
        table_ref = client.dataset(dataset_id).table(table)

        # load data
        autodetect = kwargs.pop("autodetect", True)
        job_config = bigquery.LoadJobConfig(autodetect=autodetect, **kwargs)
        if schema:
            job_config.schema = schema

        load_job = None
        try:
            load_job = client.load_table_from_uri(
                uri,
                table_ref,
                location=location,
                job_config=job_config,
            )

            # Start the job and wait for it to complete and get the result
            load_job.result()
        except Exception as exception:
            self.logger.error(exception)

            if load_job is not None and load_job.errors is not None:
                for error in load_job.errors:
                    self.logger.error(error)

            raise FAIL(exception) from exception
        # remove unpickleable attributes
        load_job._client = None
        load_job._completion_lock = None

        return load_job


class BigQueryLoadFile(Task):
    """
    Task for insert records in a Google BigQuery table via a [load
    job](https://cloud.google.com/bigquery/docs/loading-data).  Note that all of these settings
    can optionally be provided or overwritten at runtime.

    Args:
        - file (Union[str, path-like object], optional): A string or path-like object of the
            file to be loaded
        - rewind (bool, optional): if True, seek to the beginning of the file handle before
            reading the file
        - size (int, optional):  the number of bytes to read from the file handle. If size is
            None or large, resumable upload will be used. Otherwise, multipart upload will be
            used.
        - num_retries (int, optional): the number of max retries for loading the bigquery table from
            file. Defaults to `6`
        - dataset_id (str, optional): the id of a destination dataset to write the records to
        - table (str, optional): the name of a destination table to write the records to
        - project (str, optional): the project to initialize the BigQuery Client with; if not
            provided, will default to the one inferred from your credentials
        - schema (List[bigquery.SchemaField], optional): the schema to use when creating the
            table
        - location (str, optional): location of the dataset that will be queried; defaults to
            "US"
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(
        self,
        file: Union[str, Path] = None,
        rewind: bool = False,
        size: Optional[int] = None,
        num_retries: int = 6,
        dataset_id: str = None,
        table: str = None,
        project: Optional[str] = None,
        schema: Optional[List[bigquery.SchemaField]] = None,
        location: str = "US",
        **kwargs,
    ):
        self.file = file
        self.rewind = rewind
        self.size = size
        self.num_retries = num_retries
        self.dataset_id = dataset_id
        self.table = table
        self.project = project
        self.schema = schema
        self.location = location
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "file",
        "rewind",
        "size",
        "num_retries",
        "dataset_id",
        "table",
        "project",
        "location",
    )
    def run(
        self,
        file: Union[str, Path] = None,
        rewind: bool = False,
        size: int = None,
        num_retries: int = 6,
        dataset_id: str = None,
        table: str = None,
        project: Optional[str] = None,
        schema: Optional[List[bigquery.SchemaField]] = None,
        location: str = "US",
        credentials: Optional[dict] = None,
        **kwargs,
    ):
        """
        Run method for this Task. Invoked by _calling_ this Task within a Flow context, after
        initialization.

        Args:
            - file (Union[str, path-liike object], optional): A string or path-like object of
                the file to be loaded
            - rewind (bool, optional): if True, seek to the beginning of the file handle before
                reading the file
            - size (int, optional):  the number of bytes to read from the file handle. If size
                is None or large, resumable upload will be used. Otherwise, multipart upload
                will be used.
            - num_retries (int, optional): the number of max retries for loading the bigquery table from
                file. Defaults to `6`
            - dataset_id (str, optional): the id of a destination dataset to write the records
                to; if not provided here, will default to the one provided at initialization
            - table (str, optional): the name of a destination table to write the records to;
                if not provided here, will default to the one provided at initialization
            - project (str, optional): the project to initialize the BigQuery Client with; if
                not provided, will default to the one inferred from your credentials
            - schema (List[bigquery.SchemaField], optional): the schema to use when creating
                the table
            - location (str, optional): location of the dataset that will be written to;
                defaults to "US"
            - credentials (dict, optional): a JSON document containing Google Cloud
                credentials.  You should provide these at runtime with an upstream Secret task.
            - **kwargs (optional): additional kwargs to pass to the `bigquery.LoadJobConfig`;
                see the documentation here:
                https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.client.Client.html

        Raises:
            - ValueError: if all required arguments haven't been provided or file does not exist
            - IOError: if file can't be opened and read
            - ValueError: if the load job results in an error

        Returns:
            - google.cloud.bigquery.job.LoadJob: the response from `load_table_from_file`
        """
        # check for any argument inconsistencies
        if dataset_id is None or table is None:
            raise ValueError("Both dataset_id and table must be provided.")
        try:
            path = Path(file)
        except Exception as value_error:
            raise ValueError(
                "A string or path-like object must be provided."
            ) from value_error
        if not path.is_file():
            raise ValueError(f"File {path.as_posix()} does not exist.")

        # create client
        client = get_bigquery_client(project=project, credentials=credentials)

        # get table reference
        table_ref = client.dataset(dataset_id).table(table)

        # configure job
        autodetect = kwargs.pop("autodetect", True)
        job_config = bigquery.LoadJobConfig(autodetect=autodetect, **kwargs)
        if schema:
            job_config.schema = schema

        # load data
        try:
            with open(file, "rb") as file_obj:
                load_job = client.load_table_from_file(
                    file_obj,
                    table_ref,
                    rewind,
                    size,
                    num_retries,
                    location=location,
                    job_config=job_config,
                )
        except IOError as IO_error:
            raise IOError(f"Can't open and read from {path.as_posix()}.") from IO_error

        load_job.result()  # block until job is finished

        # remove unpickleable attributes
        load_job._client = None
        load_job._completion_lock = None

        return load_job


class CreateBigQueryTable(Task):
    """
    Ensures a BigQuery table exists; creates it otherwise. Note that most initialization keywords
    can optionally be provided at runtime.

    Args:
        - project (str, optional): the project to initialize the BigQuery Client with; if not
            provided, will default to the one inferred from your credentials
        - dataset (str, optional): the name of a dataset in that the table will be created
        - table (str, optional): the name of a table to create
        - schema (List[bigquery.SchemaField], optional): the schema to use when creating the table
        - clustering_fields (List[str], optional): a list of fields to cluster the table by
        - time_partitioning (bigquery.TimePartitioning, optional): a
            `bigquery.TimePartitioning` object specifying a partitioninig of the newly created
            table
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(
        self,
        project: str = None,
        dataset: str = None,
        table: str = None,
        schema: Optional[List[bigquery.SchemaField]] = None,
        clustering_fields: Optional[List[str]] = None,
        time_partitioning: Optional[bigquery.TimePartitioning] = None,
        **kwargs,
    ):
        self.project = project
        self.dataset = dataset
        self.table = table
        self.schema = schema
        self.clustering_fields = clustering_fields
        self.time_partitioning = time_partitioning
        super().__init__(**kwargs)

    @defaults_from_attrs("project", "dataset", "table", "schema")
    def run(
        self,
        project: Optional[str] = None,
        credentials: Optional[dict] = None,
        dataset: str = None,
        table: str = None,
        schema: Optional[List[bigquery.SchemaField]] = None,
    ):
        """
        Run method for this Task.  Invoked by _calling_ this Task within a Flow context, after
        initialization.

        Args:
            - project (str, optional): the project to initialize the BigQuery Client with; if
                not provided, will default to the one inferred from your credentials
            - credentials (dict, optional): a JSON document containing Google Cloud
                credentials.  You should provide these at runtime with an upstream Secret task.
                If not provided, Prefect will first check `context` for `GCP_CREDENTIALS` and
                lastly will use default Google client logic.
            - dataset (str, optional): the name of a dataset in that the table will be created
            - table (str, optional): the name of a table to create
            - schema (List[bigquery.SchemaField], optional): the schema to use when creating
                the table

        Returns:
            - None

        Raises:
            - SUCCESS: a `SUCCESS` signal if the table already exists
        """
        client = get_bigquery_client(project=project, credentials=credentials)

        try:
            dataset_ref = client.get_dataset(dataset)
        except NotFound:
            self.logger.debug("Dataset {} not found, creating...".format(dataset))
            dataset_ref = client.create_dataset(dataset)

        table_ref = dataset_ref.table(table)
        try:
            client.get_table(table_ref)
            raise SUCCESS(
                "{dataset}.{table} already exists.".format(dataset=dataset, table=table)
            )
        except NotFound:
            self.logger.debug("Table {} not found, creating...".format(table))
            table = bigquery.Table(table_ref, schema=schema)

            # partitioning
            if self.time_partitioning:
                table.time_partitioning = self.time_partitioning

            # cluster for optimal data sorting/access
            if self.clustering_fields:
                table.clustering_fields = self.clustering_fields
            client.create_table(table)



from prefect.tasks.templates.strings import StringFormatter

try:
    from prefect.tasks.templates.jinja2 import JinjaTemplate
except ImportError:
    pass

__all__ = ["JinjaTemplate", "StringFormatter"]



import firebolt.db.connection as firebolt_conn
from typing import List
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class FireboltQuery(Task):
    """
    Task for executing a query against a Firebolt database.

    Args:
        - database (str): name of the database to use.
        - username (str): username used to authenticate.
        - password (str): password used to authenticate.
        - engine_name (str): name of the engine to use.
        - query (str): query to execute against database.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor.
    """

    def __init__(
        self,
        database: str = None,
        username: str = None,
        password: str = None,
        engine_name: str = None,
        query: str = None,
        **kwargs
    ):
        self.database = database
        self.username = username
        self.password = password
        self.engine_name = engine_name
        self.query = query
        super().__init__(**kwargs)

    @defaults_from_attrs("database", "username", "password", "engine_name", "query")
    def run(
        self,
        database: str = None,
        username: str = None,
        password: str = None,
        engine_name: str = None,
        query: str = None,
    ) -> List[List]:
        """
        Task run method. Executes a query against Firebolt database.

        Args:
            - database (str): name of the database to use.
            - username (str): username used to authenticate.
            - password (str): password used to authenticate.
            - engine_name (str): name of the engine to use.
            - query (str): query to execute against database.

        Returns:
            - List[List]: output of 'cursor.fetchall()' if 'cursor.execute(query)' > 0,
                else an empty list

        Raises:
            - ValueError: if a required parameter is not supplied.
            - DatabaseError: if exception occurs when executing the query.
        """
        if not database:
            raise ValueError("A database name must be provided")
        if not username:
            raise ValueError("User name must be provided")
        if not password:
            raise ValueError("A password must be provided")
        if not engine_name:
            raise ValueError("An engine name must be provided")
        if not query:
            raise ValueError("A query string must be provided")

        # build the connection parameter dictionary
        conn_config = {
            "database": database,
            "username": username,
            "password": password,
            "engine_name": engine_name,
        }

        # connect to database, return a connection object
        conn = firebolt_conn.connect(**conn_config)
        result = []

        # execute query
        with conn:
            with conn.cursor() as cursor:
                if cursor.execute(query) > 0:
                    result = cursor.fetchall()
        return result



from prefect.tasks.secrets.base import SecretBase
from prefect.utilities.aws import get_boto_client
from prefect.utilities.tasks import defaults_from_attrs


class AWSParametersManager(SecretBase):
    """
    Task for retrieving values from AWS SSM Parameters Store and returning the parameter value.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.
    For authentication, there are two options: you can set the `AWS_CREDENTIALS` Prefect Secret
    containing your AWS access keys, which will be passed directly to the `boto3` client, or you
    can [configure your flow's runtime
    environment](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#guide-configuration)
    for `boto3`.

    Args:
        - parameter_name (str, optional): The name of the parameter to retrieve via SSM.
        - boto_kwargs (dict, optional): Additional keyword arguments to forward to the boto client.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(self, parameter_name: str = None, boto_kwargs: dict = None, **kwargs):
        self.parameter_name = parameter_name

        if boto_kwargs is None:
            self.boto_kwargs = {}
        else:
            self.boto_kwargs = boto_kwargs

        super().__init__(**kwargs)

    @defaults_from_attrs("parameter_name", "boto_kwargs")
    def run(
        self,
        parameter_name: str = None,
        credentials: str = None,
        boto_kwargs: dict = None,
    ) -> str:
        """
        Task run method.

        Args:
            - parameter_name (str): The name of the parameter to retrieve from SSM.
            - credentials (dict, optional): Your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`. If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.
            - boto_kwargs (dict, optional): Additional keyword arguments to forward to the boto client.

        Returns:
            - str: The parameter value, as a string.
        """

        if parameter_name is None:
            raise ValueError("A parameter name must be provided.")

        ssm_client = get_boto_client("ssm", credentials=credentials, **boto_kwargs)

        parameter_response = ssm_client.get_parameter(Name=parameter_name)

        parameter_value = str(parameter_response["Parameter"]["Value"])

        return parameter_value



"""
Tasks for working with files and filesystems.
"""

from .compression import Unzip, Zip
from .operations import Copy, Glob, Move, Remove

__all__ = ["Copy", "Glob", "Move", "Remove", "Unzip", "Zip"]



import os
from py2neo import Graph
from py2neo.errors import ClientError, ConnectionUnavailable

from prefect import Task
from prefect.engine.signals import FAIL
from prefect.utilities.tasks import defaults_from_attrs


class Neo4jRunCypherQueryTask(Task):
    """
    Task to run a Cypher query against Neo4j.
    More information about the Cypher query language, can be
    found at https://neo4j.com/developer/cypher/.
    This task uses `py2neo` under the hood.
    For more information, please visit
    https://py2neo.org/2021.1/index.html.

    Args:
        - server_uri (str, optional): The Neo4j URI to connect to.
            More information regarding the accepted forms for `server_uri`
            can be found at https://py2neo.org/2021.1/profiles.html.
            This parameter, if provided, takes precedence over
            `server_uri_env_var`.
        - user (str, optional): The user to use to connect
            to Neo4j.
            This parameter, if provided, takes precedence over
            `user_env_var`.
        - password (str, optional): The password to use to connect
            to Neo4j.
            This parameter, if provided, takes precedence over
            `password_env_var`.
        - db_name: (str, optional): The database name where the Cypher query
            will run.
            This parameter, if provided, takes precedence over
            `db_name_env_var`.
        - server_uri_env_var (str, optional): The name of the environment variable
            that contains the Neo4j server URI to connect to.
        - user_env_var (str, optional): The name of the environment variable
            that contains the user to use to connect to Neo4j.
        - password_env_var (str, optional): The name of the environment variable
            that contains the password to use to connect to Neo4j.
        - db_name_env_var (str, optional): The name of the environment variable
            that contains the database name where the Cypher query will run.
        - cypher_query (str, optional): The Cypher query to run.
            More information about the Cypher query language, can be
            found at https://neo4j.com/developer/cypher/.
        - return_result_as (str, optional): How to return the result.
            Accepted values are `raw`, `dataframe`.
            Defaults to `raw` (which will return a `list` of `dict`).
            Applies only when the query returned result is not empty.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    __DEFAULT_RETURN_RESULT_TYPE = "raw"
    __ACCEPTED_RETURN_RESULT_TYPES = [__DEFAULT_RETURN_RESULT_TYPE, "dataframe"]

    def __init__(
        self,
        server_uri: str = None,
        user: str = None,
        password: str = None,
        db_name: str = None,
        server_uri_env_var: str = None,
        user_env_var: str = None,
        password_env_var: str = None,
        db_name_env_var: str = None,
        cypher_query: str = None,
        return_result_as: str = __DEFAULT_RETURN_RESULT_TYPE,
        **kwargs,
    ):
        self.server_uri = server_uri
        self.user = user
        self.password = password
        self.db_name = db_name
        self.server_uri_env_var = server_uri_env_var
        self.user_env_var = user_env_var
        self.password_env_var = password_env_var
        self.db_name_env_var = db_name_env_var
        self.cypher_query = cypher_query
        self.return_result_as = return_result_as
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "server_uri",
        "user",
        "password",
        "db_name",
        "server_uri_env_var",
        "user_env_var",
        "password_env_var",
        "db_name_env_var",
        "cypher_query",
        "return_result_as",
    )
    def run(
        self,
        server_uri: str = None,
        user: str = None,
        password: str = None,
        db_name: str = None,
        server_uri_env_var: str = None,
        user_env_var: str = None,
        password_env_var: str = None,
        db_name_env_var: str = None,
        cypher_query: str = None,
        return_result_as: str = __DEFAULT_RETURN_RESULT_TYPE,
    ):
        """
        Task run method to run a Cypher query against Neo4j.

        Args:
            - server_uri (str, optional): The Neo4j URI to connect to.
                More information regarding the accepted forms for `server_uri`
                can be found at https://py2neo.org/2021.1/profiles.html.
                This parameter, if provided, takes precedence over
                `server_uri_env_var`.
            - user (str, optional): The user to use to connect
                to Neo4j.
                This parameter, if provided, takes precedence over
                `user_env_var`.
            - password (str, optional): The password to use to connect
                to Neo4j.
                This parameter, if provided, takes precedence over
                `password_env_var`.
            - db_name: (str, optional): The database name where the Cypher query
                will run.
                This parameter, if provided, takes precedence over
                `db_name_env_var`.
            - server_uri_env_var (str, optional): The name of the environment variable
                that contains the Neo4j server URI to connect to.
            - user_env_var (str, optional): The name of the environment variable
                that contains the user to use to connect to Neo4j.
            - password_env_var (str, optional): The name of the environment variable
                that contains the password to use to connect to Neo4j.
            - db_name_env_var (str, optional): The name of the environment variable
                that contains the database name where the Cypher query will run.
            - cypher_query (str, optional): The Cypher query to run.
                More information about the Cypher query language, can be
                found at https://neo4j.com/developer/cypher/.
            - return_result_as (str, optional): How to return the result.
                Accepted values are `raw`, `dataframe`.
                Defaults to `raw` (which will return a `list` of `dict`).
                Applies only when the query returned result is not empty.

        Returns:
            - `None` if the query result is empty.
            - The original result if `return_result_as` is `raw`.
            - A `pandas.DataFrame` if `return_result_as` is `dataframe`.

        Raises:
            - `ValueError` if both `server_uri` and `server_uri_env_var`
                are `None`.
            - `ValueError` if `server_uri` is `None` and `server_uri_env_var`
                is not found.
            - `ValueError` if both `user` and `user_env_var`
                are `None`.
            - `ValueError` if `user` is `None` and `user_env_var` is
                not found.
            - `ValueError` if both `password` and `password_env_var`
                are not found.
            - `ValueError` if `password` is `None` and `password_env_var`
                is not found.
            - `ValueError` if `db_name` is `None` and `db_name_env_var`
                is not found.
            - `ValueError` if `cypher_query` is `None`.
            - `ValueError` if `return_result_as` is not one of
                `raw`, `dataframe`.
            - `prefect.engine.signals.FAIL` if any error occurs while establishing
                the connection with Neo4j.
            - `prefect.engine.signals.FAIL` if any error occurs while running
                the Cypher query.
        """
        if not server_uri and not server_uri_env_var:
            msg = "Please provide either the `server_uri` or the `server_uri_env_var`."
            raise ValueError(msg)

        if not server_uri and server_uri_env_var not in os.environ.keys():
            msg = f"`{server_uri_env_var}` not found in environment variables."
            raise ValueError(msg)

        neo4j_uri = server_uri or os.environ[server_uri_env_var]

        if not user and not user_env_var:
            msg = "Please provide either the `user` or the `user_env_var`."
            raise ValueError(msg)

        if not user and user_env_var not in os.environ.keys():
            msg = f"`{user_env_var}` not found in environment variables."
            raise ValueError(msg)

        neo4j_user = user or os.environ[user_env_var]

        if not password and not password_env_var:
            msg = "Please provide either the `password` or the `password_env_var`."
            raise ValueError(msg)

        if not password and password_env_var not in os.environ.keys():
            msg = f"`{password_env_var}` not found in environment variables."
            raise ValueError(msg)

        neo4j_password = password or os.environ[password_env_var]

        neo4j_db_name = None
        if db_name:
            neo4j_db_name = db_name
        elif db_name_env_var and db_name_env_var not in os.environ.keys():
            msg = f"`{db_name_env_var}` not found in environment variables."
            raise ValueError(msg)
        elif db_name_env_var and db_name_env_var in os.environ.keys():
            neo4j_db_name = os.environ[db_name_env_var]

        if not cypher_query:
            raise ValueError("Please provide a value for `cypher_query`.")

        if return_result_as not in self.__ACCEPTED_RETURN_RESULT_TYPES:
            msg = f"Illegal value for `return_result_as`. Illegal value is: {return_result_as}."
            raise ValueError(msg)

        try:
            graph = Graph(
                profile=neo4j_uri, name=neo4j_db_name, auth=(neo4j_user, neo4j_password)
            )
        except ConnectionUnavailable as e:
            msg = f"Error while connecting to Neo4j. Exception: {str(e)}"
            raise FAIL(message=msg)

        try:
            r = graph.run(cypher_query)
        except ClientError as e:
            msg = f"Error while running Cypher query. Exception: {str(e)}"
            raise FAIL(message=msg)

        result = r.data()

        if not result:
            return None
        elif return_result_as == "dataframe":
            return r.to_data_frame()
        else:
            return result



from prefect.utilities.notifications.notifications import callback_factory
from prefect.utilities.notifications.notifications import slack_notifier
from prefect.utilities.notifications.notifications import gmail_notifier
from prefect.utilities.notifications.notifications import slack_message_formatter
from prefect.utilities.notifications.jira_notification import jira_notifier
from prefect.utilities.notifications.notifications import (
    snowflake_logger,
    snowflake_message_formatter,
)

__all__ = [
    "callback_factory",
    "gmail_notifier",
    "jira_notifier",
    "slack_message_formatter",
    "slack_notifier",
    "snowflake_logger",
    "snowflake_message_formatter",
]



"""
This module contains a collection of tasks for interacting with Azure resources.
"""

try:
    from prefect.tasks.azure.blobstorage import BlobStorageDownload, BlobStorageUpload
    from prefect.tasks.azure.cosmosdb import (
        CosmosDBCreateItem,
        CosmosDBReadItems,
        CosmosDBQueryItems,
    )
    from prefect.tasks.azure.datafactory import (
        DatafactoryCreate,
        PipelineCreate,
        PipelineRun,
    )
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.azure` requires Prefect to be installed with the "azure" extra.'
    ) from err

__all__ = [
    "BlobStorageDownload",
    "BlobStorageUpload",
    "CosmosDBCreateItem",
    "CosmosDBQueryItems",
    "CosmosDBReadItems",
    "DatafactoryCreate",
    "PipelineCreate",
    "PipelineRun",
]



import base64
import io
import json
import importlib

from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple

import cloudpickle
import pendulum

if TYPE_CHECKING:
    import pandas as pd

__all__ = (
    "Serializer",
    "PickleSerializer",
    "JSONSerializer",
    "DateTimeSerializer",
    "PandasSerializer",
    "CompressedSerializer",
)


class Serializer:
    """
    Serializers are used by Results to handle the transformation of Python
    objects to and from bytes.

    Subclasses should implement `serialize` and `deserialize`.
    """

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other)

    def serialize(self, value: Any) -> bytes:
        """
        Serialize an object to bytes.

        Args:
            - value (Any): the value to serialize

        Returns:
            - bytes: the serialized value
        """
        raise NotImplementedError

    def deserialize(self, value: bytes) -> Any:
        """
        Deserialize an object from bytes.

        Args:
            - value (bytes): the value to deserialize

        Returns:
            - Any: the deserialized value
        """
        raise NotImplementedError


class PickleSerializer(Serializer):
    """A `Serializer` that uses cloudpickle to serialize Python objects."""

    def serialize(self, value: Any) -> bytes:
        """
        Serialize an object to bytes using cloudpickle.

        Args:
            - value (Any): the value to serialize

        Returns:
            - bytes: the serialized value
        """
        return cloudpickle.dumps(value)

    def deserialize(self, value: bytes) -> Any:
        """
        Deserialize an object from bytes using cloudpickle.

        Args:
            - value (bytes): the value to deserialize

        Returns:
            - Any: the deserialized value
        """
        try:
            return cloudpickle.loads(value)
        except Exception as exc:
            try:
                # old versions of Core encoded pickles with base64
                return cloudpickle.loads(base64.b64decode(value))
            except Exception as e:
                # if there's an error with the backwards-compatible step,
                # reraise the original exception
                raise exc from e


class JSONSerializer(Serializer):
    """A Serializer that uses JSON to serialize objects"""

    def serialize(self, value: Any) -> bytes:
        """
        Serialize an object to JSON

        Args:
            - value (Any): the value to serialize

        Returns:
            - bytes: the serialized value
        """
        return json.dumps(value).encode()

    def deserialize(self, value: bytes) -> Any:
        """
        Deserialize an object from JSON

        Args:
            - value (bytes): the value to deserialize

        Returns:
            - Any: the deserialized value
        """
        return json.loads(value)


class DateTimeSerializer(Serializer):
    """A Serializer for working with human-readable datetimes"""

    def serialize(self, value: Any) -> bytes:
        """
        Serialize a datetime to human-readable bytes

        Args:
            - value (Any): the value to serialize

        Returns:
            - bytes: the serialized value
        """
        return pendulum.instance(value).to_iso8601_string().encode()

    def deserialize(self, value: bytes) -> Any:
        """
        Deserialize an datetime from human-readable bytes

        Args:
            - value (bytes): the value to deserialize

        Returns:
            - Any: the deserialized value
        """
        return pendulum.parse(value.decode())


class PandasSerializer(Serializer):
    """A Serializer for Pandas DataFrames.

    Args:
        - file_type (str): The type you want the resulting file to be
            saved as, e.g. "csv" or "parquet". Must match a type used
            in a `DataFrame.to_` method and a `pd.read_` function.
        - deserialize_kwargs (dict, optional): Keyword arguments to pass to the
            deserialization method.
        - serialize_kwargs (dict, optional): Keyword arguments to pass to the
            serialization method.
    """

    def __init__(
        self,
        file_type: str,
        deserialize_kwargs: dict = None,
        serialize_kwargs: dict = None,
    ) -> None:
        self.file_type = file_type

        # Fails fast if user specifies a format that Pandas can't deal with.
        self._get_deserialize_method()
        self._get_serialize_method()

        self.deserialize_kwargs = (
            {} if deserialize_kwargs is None else deserialize_kwargs
        )
        self.serialize_kwargs = {} if serialize_kwargs is None else serialize_kwargs

    def serialize(self, value: "pd.DataFrame") -> bytes:  # noqa: F821
        """
        Serialize a Pandas DataFrame to bytes.

        Args:
            - value (DataFrame): the DataFrame to serialize

        Returns:
            - bytes: the serialized value
        """
        serialization_method = self._get_serialize_method(dataframe=value)
        buffer = io.BytesIO()
        try:
            serialization_method(buffer, **self.serialize_kwargs)
            return buffer.getvalue()
        except TypeError:
            # there are some weird bugs with several of the Pandas serialization
            # methods when trying to serialize to bytes directly. This is a
            # workaround. See https://github.com/pandas-dev/pandas/pull/35129
            string_buffer = io.StringIO()
            serialization_method(string_buffer, **self.serialize_kwargs)
            return string_buffer.getvalue().encode()

    def deserialize(self, value: bytes) -> "pd.DataFrame":  # noqa: F821
        """
        Deserialize an object to a Pandas DataFrame

        Args:
            - value (bytes): the value to deserialize

        Returns:
            - DataFrame: the deserialized DataFrame
        """
        deserialization_method = self._get_deserialize_method()
        buffer = io.BytesIO(value)
        deserialized_data = deserialization_method(buffer, **self.deserialize_kwargs)
        return deserialized_data

    def __eq__(self, other: Any) -> bool:
        if type(self) == type(other):
            return (
                self.file_type == other.file_type
                and self.serialize_kwargs == other.serialize_kwargs
                and self.deserialize_kwargs == other.deserialize_kwargs
            )
        return False

    # _get_read_method and _get_write_method are constructed as they are both to
    # limit copy/paste but also to make it easier for potential future extension to serialization
    # methods that do not map to the "to_{}/read_{}" interface.
    def _get_deserialize_method(self) -> Callable:
        import pandas as pd

        try:
            return getattr(pd, "read_{}".format(self.file_type))
        except AttributeError as exc:
            raise ValueError(
                "Could not find deserialization methods for {}".format(self.file_type)
            ) from exc

    def _get_serialize_method(self, dataframe: "pd.DataFrame" = None) -> Callable:
        import pandas as pd

        if dataframe is None:
            # If you just want to test if the method exists, create an empty dataframe
            dataframe = pd.DataFrame()
        try:
            return getattr(dataframe, "to_{}".format(self.file_type))
        except AttributeError as exc:
            raise ValueError(
                "Could not find serialization methods for {}".format(self.file_type)
            ) from exc


class CompressedSerializer(Serializer):
    """
    A Serializer that wraps another Serializer and a compression function to serialize
    Python objects with compression.

    Args:
        - serializer (Serializer): the serializer that this serializer wraps
        - format (str): name of the compression format library. Typically one of the
            python standard compression libraries: bz2, gzip, lzma, or zlib. Attempts
            to import the given format's  module and retrieves the compress/decompress
            functions.
        - compress (Callable[..., bytes]): the custom compression function
        - decompress (Callable[..., bytes]): the custom decompression function
        - compress_kwargs (Dict[str, Any]): keyword arguments to be passed to the
            compression function
        - decompress_kwargs (Dict[str, Any]): keyword arguments to be passed to the
            decompression function
    """

    def __init__(
        self,
        serializer: Serializer,
        format: str = None,
        compress: Callable[..., bytes] = None,
        decompress: Callable[..., bytes] = None,
        compress_kwargs: Dict[str, Any] = None,
        decompress_kwargs: Dict[str, Any] = None,
    ):
        self._serializer = serializer

        if format and (compress or decompress):
            raise ValueError(
                "You must specify either `format` or `compress`/`decompress`, "
                "but not both."
            )
        elif format:
            self._compress, self._decompress = self.compression_from_lib(format)
        elif compress and decompress:
            self._compress = compress
            self._decompress = decompress
        else:
            raise ValueError(
                "You must specify either `format` or `compress`/`decompress`."
            )

        self._compress_kwargs = compress_kwargs or {}
        self._decompress_kwargs = decompress_kwargs or {}

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self._serializer == other._serializer
            and self._compress == other._compress
            and self._decompress == other._decompress
            and self._compress_kwargs == other._compress_kwargs
            and self._decompress_kwargs == other._decompress_kwargs
        )

    def serialize(self, value: Any) -> bytes:
        """
        Serialize an object to compressed bytes.

        Args:
            - value (Any): the value to serialize

        Returns:
            - bytes: the compressed serialized value
        """
        return self._compress(
            self._serializer.serialize(value), **self._compress_kwargs
        )

    def deserialize(self, value: bytes) -> Any:
        """
        Deserialize an object from compressed bytes.

        Args:
            - value (bytes): the compressed value to deserialize

        Returns:
            - Any: the deserialized value
        """
        return self._serializer.deserialize(
            self._decompress(value, **self._decompress_kwargs)
        )

    @staticmethod
    def compression_from_lib(
        compression_format: str,
    ) -> Tuple[Callable[..., bytes], Callable[..., bytes]]:
        """
        Attempt to pull a compression format from a library. Typically one of
        "lzma", "gzip", "zlib", "bz2"

        Args:
            - compression_format: The compression format/library to load

        Returns:
            A tuple of functions for compression and decompression
        """
        common_libs = {"lzma", "gzip", "zlib", "bz2"}

        # Don't suggest them a format they've just requested
        common_libs.discard(compression_format)

        try:
            module = importlib.import_module(compression_format)
        except ImportError as exc:
            raise ImportError(
                f"Compression module for {compression_format!r} is not installed. "
                f"Did you mean to use one of {common_libs}?"
            ) from exc

        try:
            funcs = (module.compress, module.decompress)  # type: ignore
        except AttributeError as exc:
            raise ValueError(
                f"Given compression format {compression_format!r} module does not have "
                f"'compress' and 'decompress' attributes. Pass these functions "
                f"manually instead if you intend to use a non-standard library. "
                f"Otherwise, use one of the common compression libraries: {common_libs}"
            ) from exc

        return funcs



"""
The functions here have been moved to `prefect.backend.artifacts`
"""
# flake8: noqa
import warnings
from typing import Optional

from prefect.backend import (
    create_link_artifact,
    create_markdown_artifact,
    delete_artifact as delete_artifact_new,
    update_link_artifact,
    update_markdown_artifact,
)


def create_link(link: str) -> Optional[str]:
    """
    Create a link artifact

    Args:
        - link (str): the link to post

    Returns:
        - str: the task run artifact ID
    """
    warnings.warn(
        "`prefect.artifacts.create_link` has been moved to `prefect.backend.create_link_artifact`. "
        "Please update your imports. This import path will be removed in 1.0.0."
    )
    return create_link_artifact(link)


def update_link(task_run_artifact_id: str, link: str) -> None:
    """
    Update an existing link artifact. This function will replace the current link
    artifact with the new link provided.

    Args:
        - task_run_artifact_id (str): the ID of an existing task run artifact
        - link (str): the new link to update the artifact with
    """
    warnings.warn(
        "`prefect.artifacts.update_link` has been moved to `prefect.backend.update_link_artifact`. "
        "Please update your imports. This import path will be removed in 1.0.0."
    )
    return update_link_artifact(task_run_artifact_id, link)


def create_markdown(markdown: str) -> Optional[str]:
    """
    Create a markdown artifact

    Args:
        - markdown (str): the markdown to post

    Returns:
        - str: the task run artifact ID
    """
    warnings.warn(
        "`prefect.artifacts.create_markdown` has been moved to `prefect.backend.create_markdown_artifact`. "
        "Please update your imports. This import path will be removed in 1.0.0."
    )
    return create_markdown_artifact(markdown)


def update_markdown(task_run_artifact_id: str, markdown: str) -> None:
    """
    Update an existing markdown artifact. This function will replace the current markdown
    artifact with the new markdown provided.

    Args:
        - task_run_artifact_id (str): the ID of an existing task run artifact
        - markdown (str): the new markdown to update the artifact with
    """
    warnings.warn(
        "`prefect.artifacts.update_markdown` has been moved to `prefect.backend.update_markdown_artifact`. "
        "Please update your imports. This import path will be removed in 1.0.0."
    )
    return update_markdown_artifact(task_run_artifact_id, markdown)


def delete_artifact(task_run_artifact_id: str) -> None:
    """
    Delete an existing artifact

    Args:
        - task_run_artifact_id (str): the ID of an existing task run artifact
    """
    warnings.warn(
        "`prefect.artifacts.delete_artifact` has been moved to `prefect.backend.delete_artifact`. "
        "Please update your imports. This import path will be removed in 1.0.0."
    )
    return delete_artifact_new(task_run_artifact_id)
